"""Various conversion datapipes"""
from ocf_datapipes.convert.numpy.batch.gsp import (
    ConvertGSPToNumpyBatchIterDataPipe as ConvertGSPToNumpyBatch,
)

from .coordinates import ConvertGeostationaryToLatLonIterDataPipe as ConvertGeostationaryToLatLon
from .coordinates import ConvertLatLonToOSGBIterDataPipe as ConvertLatLonToOSGB
from .coordinates import ConvertOSGBToLatLonIterDataPipe as ConvertOSGBToLatLon
from .numpy.gsp import ConvertGSPToNumpyIterDataPipe as ConvertGSPToNumpy
from .nwp import ConvertNWPToNumpyBatchIterDataPipe as ConvertNWPToNumpyBatch
from .pv import ConvertPVToNumpyBatchIterDataPipe as ConvertPVToNumpyBatch
from .satellite import ConvertSatelliteToNumpyBatchIterDataPipe as ConvertSatelliteToNumpyBatch



from ocf_datapipes.select import DropGSP, LocationPicker
from ocf_datapipes.transform.xarray import ConvertToNWPTargetTime, CreatePVImage, PreProcessMetNet


def test_metnet_preprocess_no_sun(sat_datapipe, gsp_datapipe):
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])
    gsp_datapipe = LocationPicker(gsp_datapipe)
    datapipe = PreProcessMetNet(
        [sat_datapipe],
        location_datapipe=gsp_datapipe,
        center_width=100_000,
        center_height=100_000,
        context_height=1_000_000,
        context_width=1_000_000,
        output_width_pixels=100,
        output_height_pixels=100,
        add_sun_features=False,
    )
    data = next(iter(datapipe))
    print(data.shape)


def test_metnet_preprocess(sat_datapipe, gsp_datapipe):
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])
    gsp_datapipe = LocationPicker(gsp_datapipe)
    datapipe = PreProcessMetNet(
        [sat_datapipe],
        location_datapipe=gsp_datapipe,
        center_width=100_000,
        center_height=100_000,
        context_height=1_000_000,
        context_width=1_000_000,
        output_width_pixels=100,
        output_height_pixels=100,
        add_sun_features=True,
    )
    data = next(iter(datapipe))
    print(data.shape)


def test_metnet_preprocess_both_sat(sat_datapipe, sat_hrv_datapipe, gsp_datapipe):
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])
    gsp_datapipe = LocationPicker(gsp_datapipe)
    datapipe = PreProcessMetNet(
        [sat_datapipe, sat_hrv_datapipe],
        location_datapipe=gsp_datapipe,
        center_width=100_000,
        center_height=100_000,
        context_height=1_000_000,
        context_width=1_000_000,
        output_width_pixels=100,
        output_height_pixels=100,
        add_sun_features=False,
    )
    data = next(iter(datapipe))
    print(data.shape)


def test_metnet_preprocess_both_sat_other_order(sat_datapipe, sat_hrv_datapipe, gsp_datapipe):
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])
    gsp_datapipe = LocationPicker(gsp_datapipe)
    datapipe = PreProcessMetNet(
        [sat_hrv_datapipe, sat_datapipe],
        location_datapipe=gsp_datapipe,
        center_width=100_000,
        center_height=100_000,
        context_height=1_000_000,
        context_width=1_000_000,
        output_width_pixels=100,
        output_height_pixels=100,
        add_sun_features=True,
    )
    data = next(iter(datapipe))
    print(data.shape)


def test_metnet_preprocess_both_sat_pv(
    sat_datapipe, sat_hrv_datapipe, gsp_datapipe, passiv_datapipe
):
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])
    gsp_datapipe = LocationPicker(gsp_datapipe)
    sat_datapipe, image_datapipe = sat_datapipe.fork(2)
    passiv_datapipe = CreatePVImage(passiv_datapipe, image_datapipe=image_datapipe, normalize=True)
    datapipe = PreProcessMetNet(
        [sat_hrv_datapipe, sat_datapipe, passiv_datapipe],
        location_datapipe=gsp_datapipe,
        center_width=100_000,
        center_height=100_000,
        context_height=1_000_000,
        context_width=1_000_000,
        output_width_pixels=100,
        output_height_pixels=100,
        add_sun_features=True,
    )
    data = next(iter(datapipe))
    assert data.shape == (289, 14, 100, 100)


def test_metnet_preprocess_sat_hrv_pv_nwp(
    sat_datapipe, sat_hrv_datapipe, gsp_datapipe, passiv_datapipe, nwp_datapipe
):
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])
    gsp_datapipe = LocationPicker(gsp_datapipe)
    sat_datapipe, image_datapipe = sat_datapipe.fork(2)
    passiv_datapipe = CreatePVImage(passiv_datapipe, image_datapipe=image_datapipe, normalize=True)
    datapipe = PreProcessMetNet(
        [sat_hrv_datapipe, sat_datapipe, passiv_datapipe],
        location_datapipe=gsp_datapipe,
        center_width=100_000,
        center_height=100_000,
        context_height=1_000_000,
        context_width=1_000_000,
        output_width_pixels=100,
        output_height_pixels=100,
        add_sun_features=True,
    )
    data = next(iter(datapipe))
    assert data.shape == (289, 14, 100, 100)



"""Merge individual examples into a batch"""
import logging

from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import NumpyBatch
from ocf_datapipes.utils.utils import stack_np_examples_into_batch

logger = logging.getLogger(__name__)


@functional_datapipe("merge_numpy_examples_to_batch")
class MergeNumpyExamplesToBatchIterDataPipe(IterDataPipe):
    """Merge individual examples into a batch"""

    def __init__(self, source_datapipe: IterDataPipe, n_examples_per_batch: int):
        """
        Merge individual examples into a batch

        Args:
            source_datapipe: Datapipe of NumpyBatch data
            n_examples_per_batch: Number of examples per batch
        """
        self.source_datapipe = source_datapipe
        self.n_examples_per_batch = n_examples_per_batch

    def __iter__(self) -> NumpyBatch:
        """Merge individual examples into a batch"""
        np_examples = []
        logger.debug("Merging numpy batch")
        for np_batch in self.source_datapipe:
            np_examples.append(np_batch)
            if len(np_examples) == self.n_examples_per_batch:
                yield stack_np_examples_into_batch(np_examples)
                np_examples = []



"""Constants and Enums."""

from enum import Enum, auto
from typing import Optional, Union

import numpy as np
import xarray as xr
from pydantic import BaseModel

PV_TIME_AXIS = 1
PV_SYSTEM_AXIS = 2

Y_OSGB_MEAN = 357021.38
Y_OSGB_STD = 612920.2
X_OSGB_MEAN = 187459.94
X_OSGB_STD = 622805.44

SATELLITE_SPACER_LEN = 17  # Patch of 4x4 + 1 for surface height.
PV_SPACER_LEN = 18  # 16 for embedding dim + 1 for marker + 1 for history

PV_SYSTEM_ID: str = "pv_system_id"
PV_SYSTEM_ROW_NUMBER = "pv_system_row_number"
PV_SYSTEM_X_COORDS = "pv_system_x_coords"
PV_SYSTEM_Y_COORDS = "pv_system_y_coords"

SUN_AZIMUTH_ANGLE = "sun_azimuth_angle"
SUN_ELEVATION_ANGLE = "sun_elevation_angle"

PV_YIELD = "pv_yield"
PV_DATETIME_INDEX = "pv_datetime_index"
DEFAULT_N_PV_SYSTEMS_PER_EXAMPLE = 2048
GSP_ID: str = "gsp_id"
GSP_YIELD = "gsp_yield"
GSP_X_COORDS = "gsp_x_coords"
GSP_Y_COORDS = "gsp_y_coords"
GSP_DATETIME_INDEX = "gsp_datetime_index"
N_GSPS = 317

DEFAULT_N_GSP_PER_EXAMPLE = 32
OBJECT_AT_CENTER = "object_at_center"
DATETIME_FEATURE_NAMES = (
    "hour_of_day_sin",
    "hour_of_day_cos",
    "day_of_year_sin",
    "day_of_year_cos",
)
SATELLITE_DATA = "sat_data"
SATELLITE_Y_COORDS = "sat_y_coords"
SATELLITE_X_COORDS = "sat_x_coords"
SATELLITE_DATETIME_INDEX = "sat_datetime_index"
NWP_TARGET_TIME = "nwp_target_time"
NWP_DATA = "nwp"
NWP_X_COORDS = "nwp_x_coords"
NWP_Y_COORDS = "nwp_y_coords"
X_CENTERS_OSGB = "x_centers_osgb"
Y_CENTERS_OSGB = "y_centers_osgb"
TOPOGRAPHIC_DATA = "topo_data"
TOPOGRAPHIC_X_COORDS = "topo_x_coords"
TOPOGRAPHIC_Y_COORDS = "topo_y_coords"

# "Safe" default NWP variable names:
NWP_VARIABLE_NAMES = (
    "t",
    "dswrf",
    "prate",
    "r",
    "sde",
    "si10",
    "vis",
    "lcc",
    "mcc",
    "hcc",
)

# A complete set of NWP variable names.  Not all are currently used.
FULL_NWP_VARIABLE_NAMES = (
    "cdcb",
    "lcc",
    "mcc",
    "hcc",
    "sde",
    "hcct",
    "dswrf",
    "dlwrf",
    "h",
    "t",
    "r",
    "dpt",
    "vis",
    "si10",
    "wdir10",
    "prmsl",
    "prate",
)

SAT_VARIABLE_NAMES = (
    "HRV",
    "IR_016",
    "IR_039",
    "IR_087",
    "IR_097",
    "IR_108",
    "IR_120",
    "IR_134",
    "VIS006",
    "VIS008",
    "WV_062",
    "WV_073",
)

DEFAULT_REQUIRED_KEYS = [
    NWP_DATA,
    NWP_X_COORDS,
    NWP_Y_COORDS,
    SATELLITE_DATA,
    SATELLITE_X_COORDS,
    SATELLITE_Y_COORDS,
    PV_YIELD,
    PV_SYSTEM_ID,
    PV_SYSTEM_ROW_NUMBER,
    PV_SYSTEM_X_COORDS,
    PV_SYSTEM_Y_COORDS,
    X_CENTERS_OSGB,
    Y_CENTERS_OSGB,
    GSP_ID,
    GSP_YIELD,
    GSP_X_COORDS,
    GSP_Y_COORDS,
    GSP_DATETIME_INDEX,
    TOPOGRAPHIC_DATA,
    TOPOGRAPHIC_Y_COORDS,
    TOPOGRAPHIC_X_COORDS,
] + list(DATETIME_FEATURE_NAMES)
T0_DT = "t0_dt"


SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME = (
    "spatial_and_temporal_locations_of_each_example.csv"
)

LOG_LEVELS = ("DEBUG", "INFO", "WARNING", "ERROR")


class Location(BaseModel):
    """Represent a spatial location."""

    x: float
    y: float
    id: Optional[int]


class BatchKey(Enum):
    """The names of the different elements of each batch.

    This is also where we document the exact shape of each element.

    Each `DataSource` may be split into several different BatchKey elements. For example, the
    PV PreparedDataSource yields `pv` and `pv_system_row_number` BatchKeys.
    """

    # -------------- HRVSATELLITE -----------------------------------
    # shape: (batch_size, time, channels, y, x)
    #
    # Or, if the imagery has been patched,
    # shape: (batch_size, time, channels, y, x, n_pixels_per_patch) where n_pixels_per_patch
    # is the *total* number of pixels,
    # i.e. n_pixels_per_patch_along_height * n_pixels_per_patch_along_width.
    hrvsatellite_actual = auto()
    hrvsatellite_predicted = auto()  # shape: batch_size, time, y, x
    hrvsatellite_t0_idx = auto()  # shape: scalar

    # HRV satellite coordinates:
    hrvsatellite_y_osgb = auto()  # shape: (batch_size, y, x)
    hrvsatellite_x_osgb = auto()  # shape: (batch_size, y, x)
    hrvsatellite_y_geostationary = auto()  # shape: (batch_size, y)
    hrvsatellite_x_geostationary = auto()  # shape: (batch_size, x)
    #: Time is seconds since UNIX epoch (1970-01-01). Shape: (batch_size, n_timesteps)
    hrvsatellite_time_utc = auto()
    # Added by np_batch_processor.Topography:
    hrvsatellite_surface_height = auto()  # The surface height at each pixel. (batch_size, y, x)

    # HRV satellite Fourier coordinates:
    # Spatial coordinates. Shape: (batch_size, y, x, n_fourier_features_per_dim)
    hrvsatellite_y_osgb_fourier = auto()
    hrvsatellite_x_osgb_fourier = auto()
    #: Time shape: (batch_size, n_timesteps, n_fourier_features_per_dim)
    hrvsatellite_time_utc_fourier = auto()
    hrvsatellite_time_utc_fourier_t0 = auto()

    # -------------- NWP --------------------------------------------
    nwp = auto()  # shape: (batch_size, target_time_utc, channel, y_osgb, x_osgb)
    nwp_t0_idx = auto()  # shape: scalar
    nwp_target_time_utc = auto()  # shape: (batch_size, target_time_utc)
    nwp_init_time_utc = auto()  # shape: (batch_size, target_time_utc)
    nwp_step = auto()  # Int. Number of hours. shape: (batch_size, target_time_utc)
    nwp_y_osgb = auto()  # shape: (batch_size, y_osgb)
    nwp_x_osgb = auto()  # shape: (batch_size, x_osgb)
    nwp_channel_names = auto()  # shape: (channel,)

    # NWP Fourier features:
    nwp_target_time_utc_fourier = auto()
    nwp_init_time_utc_fourier = auto()
    nwp_y_osgb_fourier = auto()
    nwp_x_osgb_fourier = auto()

    # -------------- PV ---------------------------------------------
    pv = auto()  # shape: (batch_size, time, n_pv_systems)
    pv_t0_idx = auto()  # shape: scalar
    pv_system_row_number = auto()  # shape: (batch_size, n_pv_systems)
    pv_id = auto()  # shape: (batch_size, n_pv_systems)
    # PV AC system capacity in watts peak.
    # Warning: In v15, pv_capacity_watt_power is sometimes 0. This will be fixed in
    # https://github.com/openclimatefix/nowcasting_dataset/issues/622
    pv_capacity_watt_power = auto()  # shape: (batch_size, n_pv_systems)
    #: pv_mask is True for good PV systems in each example.
    # The RawPVDataSource doesn't use pv_mask. Instead is sets missing PV systems to NaN
    # across all PV batch keys.
    pv_mask = auto()  # shape: (batch_size, n_pv_systems)

    # PV coordinates:
    # Each has shape: (batch_size, n_pv_systems), will be NaN for missing PV systems.
    pv_y_osgb = auto()
    pv_x_osgb = auto()
    pv_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01).

    # PV Fourier coordinates:
    # Each has shape: (batch_size, n_pv_systems, n_fourier_features_per_dim),
    # and will be NaN for missing PV systems.
    pv_y_osgb_fourier = auto()
    pv_x_osgb_fourier = auto()
    pv_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)
    pv_time_utc_fourier_t0 = auto()  # Added by SaveT0Time. Shape: (batch_size, n_fourier_features)

    # -------------- GSP --------------------------------------------
    gsp = auto()  # shape: (batch_size, time, 1)  (the RawGSPDataSource include a '1',
    # not sure if the prepared dataset does!)
    gsp_t0_idx = auto()  # shape: scalar
    gsp_id = auto()  # shape: (batch_size)

    # GSP coordinates:
    # Each has shape: (batch_size). No NaNs.
    gsp_y_osgb = auto()
    gsp_x_osgb = auto()
    gsp_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01). (batch_size, time)
    gsp_capacity_megawatt_power = auto()  # (batch_size)

    # GSP Fourier coordinates:
    # Each has shape: (batch_size, 1, n_fourier_features_per_dim),
    # no NaNs.
    gsp_y_osgb_fourier = auto()
    gsp_x_osgb_fourier = auto()
    gsp_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)
    gsp_time_utc_fourier_t0 = auto()  # Added by SaveT0Time. Shape: (batch_size, n_fourier_features)

    # -------------- GSP5Min ----------------------------------------
    # Not used by the Raw data pipeline!
    gsp_5_min = auto()  # shape: (batch_size, time)
    gsp_5_min_time_utc = auto()  # shape: (batch_size, time)
    gsp_5_min_time_utc_fourier = auto()  # shape: (batch_size, time, n_fourier_features)
    gsp_5_min_t0_idx = auto()

    # -------------- SUN --------------------------------------------
    # Solar position at every timestep. shape = (batch_size, n_timesteps)
    # The solar position data comes from two alternative sources: either the Sun pre-prepared
    # batches, or the SunPosition np_batch_processor for the RawDataset.
    hrvsatellite_solar_azimuth = auto()
    hrvsatellite_solar_elevation = auto()
    satellite_solar_azimuth = auto()
    satellite_solar_elevation = auto()
    gsp_solar_azimuth = auto()
    gsp_solar_elevation = auto()
    gsp_5_min_solar_azimuth = auto()
    gsp_5_min_solar_elevation = auto()
    pv_solar_azimuth = auto()
    pv_solar_elevation = auto()
    nwp_target_time_solar_azimuth = auto()
    nwp_target_time_solar_elevation = auto()

    # Solar position at the centre of the HRV image at t0
    # (from `power_perceiver.np_batch_processor.SunPosition`)
    # shape = (example,)
    # Not used in Raw data pipeline.
    solar_azimuth_at_t0 = auto()
    solar_elevation_at_t0 = auto()

    # ------------- REQUESTED NUMBER OF TIMESTEPS ---------------------
    # Added by `ReduceNumTimesteps`. Gives the indicies of the randomly selected timesteps.
    # Not used in the Raw data pipeline.
    requested_timesteps = auto()  # shape: (n_requested_timesteps)

    # -------------- SATELLITE -----------------------------------
    # shape: (batch_size, time, channels, y, x)
    #
    # Or, if the imagery has been patched,
    # shape: (batch_size, time, channels, y, x, n_pixels_per_patch) where n_pixels_per_patch
    # is the *total* number of pixels,
    # i.e. n_pixels_per_patch_along_height * n_pixels_per_patch_along_width.
    satellite_actual = auto()
    satellite_predicted = auto()  # shape: batch_size, time, y, x
    satellite_t0_idx = auto()  # shape: scalar

    # HRV satellite coordinates:
    satellite_y_osgb = auto()  # shape: (batch_size, y, x)
    satellite_x_osgb = auto()  # shape: (batch_size, y, x)
    satellite_y_geostationary = auto()  # shape: (batch_size, y)
    satellite_x_geostationary = auto()  # shape: (batch_size, x)
    #: Time is seconds since UNIX epoch (1970-01-01). Shape: (batch_size, n_timesteps)
    satellite_time_utc = auto()
    # Added by np_batch_processor.Topography:
    satellite_surface_height = auto()  # The surface height at each pixel. (batch_size, y, x)

    # HRV satellite Fourier coordinates:
    # Spatial coordinates. Shape: (batch_size, y, x, n_fourier_features_per_dim)
    satellite_y_osgb_fourier = auto()
    satellite_x_osgb_fourier = auto()
    #: Time shape: (batch_size, n_timesteps, n_fourier_features_per_dim)
    satellite_time_utc_fourier = auto()
    satellite_time_utc_fourier_t0 = auto()


NumpyBatch = dict[BatchKey, np.ndarray]

XarrayBatch = dict[BatchKey, Union[xr.DataArray, xr.Dataset]]

# Means and std computed with
# nowcasting_dataset/scripts/compute_stats_from_batches.py
# using v15 training batches on 2021-11-24.
NWP_MEAN = {
    "t": 285.7799539185846,
    "dswrf": 294.6696933986283,
    "prate": 3.6078121378638696e-05,
    "r": 75.57106712435926,
    "sde": 0.0024915961594965614,
    "si10": 4.931356852411006,
    "vis": 22321.762918384553,
    "lcc": 47.90454236572895,
    "mcc": 44.22781694449808,
    "hcc": 32.87577371914454,
}

NWP_STD = {
    "t": 5.017000766747606,
    "dswrf": 233.1834250473355,
    "prate": 0.00021690701537950742,
    "r": 15.705370079694358,
    "sde": 0.07560040052148084,
    "si10": 2.664583614352396,
    "vis": 12963.802514945439,
    "lcc": 40.06675870700349,
    "mcc": 41.927221148316384,
    "hcc": 39.05157559763763,
}

NWP_CHANNEL_NAMES = tuple(NWP_STD.keys())


def _to_data_array(d):
    return xr.DataArray(
        [d[key] for key in NWP_CHANNEL_NAMES], coords={"channel": list(NWP_CHANNEL_NAMES)}
    ).astype(np.float32)


NWP_MEAN = _to_data_array(NWP_MEAN)
NWP_STD = _to_data_array(NWP_STD)

SAT_MEAN = {
    "HRV": 236.13257536395903,
    "IR_016": 291.61620182554185,
    "IR_039": 858.8040610176552,
    "IR_087": 738.3103442750336,
    "IR_097": 773.0910794778366,
    "IR_108": 607.5318145165666,
    "IR_120": 860.6716261423857,
    "IR_134": 925.0477987594331,
    "VIS006": 228.02134593063957,
    "VIS008": 257.56333202381205,
    "WV_062": 633.5975770915588,
    "WV_073": 543.4963868823854,
}

SAT_STD = {
    "HRV": 935.9717382401759,
    "IR_016": 172.01044433112992,
    "IR_039": 96.53756504807913,
    "IR_087": 96.21369354283686,
    "IR_097": 86.72892737648276,
    "IR_108": 156.20651744208888,
    "IR_120": 104.35287930753246,
    "IR_134": 104.36462050405994,
    "VIS006": 150.2399269307514,
    "VIS008": 152.16086321818398,
    "WV_062": 111.8514878214775,
    "WV_073": 106.8855172848904,
}


def _to_data_array(d):
    return xr.DataArray(
        [d[key] for key in SAT_VARIABLE_NAMES], coords={"channel": list(SAT_VARIABLE_NAMES)}
    ).astype(np.float32)


SAT_MEAN_DA = _to_data_array(SAT_MEAN)
SAT_STD_DA = _to_data_array(SAT_STD)



""" Function to split datasets up """

import logging
from collections import namedtuple
from enum import Enum
from typing import List, Optional, Tuple, Union

import pandas as pd

from ocf_datapipes.utils.split.method import split_by_dates, split_method
from ocf_datapipes.utils.split.model import (
    TrainValidationTestSpecific,
    default_train_test_validation_specific,
)

logger = logging.getLogger(__name__)


class SplitMethod(Enum):
    """Different split methods"""

    DATE = "date"
    DAY = "day"
    DAY_RANDOM = "day_random"
    DAY_SPECIFIC = "day_specific"
    WEEK = "week"
    WEEK_RANDOM = "week_random"
    YEAR_SPECIFIC = "year_specific"
    SAME = "same"
    DAY_RANDOM_TEST_YEAR = "day_random_test_year"
    DAY_RANDOM_TEST_DATE = "day_random_test_date"


class SplitName(Enum):
    """The name for each data split."""

    TRAIN = "train"
    VALIDATION = "validation"
    TEST = "test"


# Create a namedtuple for storing split t0 datetimes.
SplitDateTimes = namedtuple(
    typename="SplitDateTimes",
    field_names=[SplitName.TRAIN.value, SplitName.VALIDATION.value, SplitName.TEST.value],
)


def split_data(
    datetimes: Union[List[pd.Timestamp], pd.DatetimeIndex],
    method: SplitMethod,
    train_test_validation_split: Tuple[int, int, int] = (3, 1, 1),
    train_test_validation_specific: TrainValidationTestSpecific = (
        default_train_test_validation_specific
    ),
    train_validation_test_datetime_split: Optional[List[pd.Timestamp]] = None,
    seed: int = 1234,
) -> SplitDateTimes:
    """
    Split the date using various different methods

    Args:
        datetimes: The datetimes to be split
        method: the method to be used
        train_test_validation_split: ratios of how the split is made
        seed: random seed used to permutate the data for the 'random' method
        train_test_validation_specific: pydandic class of 'train', 'validation' and 'test'.
            These specify which data goes into which dataset.
        train_validation_test_datetime_split: split train, validation based on specific dates.

    Returns: train, validation and test dataset
    """
    logger.info(f"Splitting data with method {method}")

    datetimes = pd.DatetimeIndex(datetimes)

    if method == SplitMethod.SAME:
        train_datetimes = datetimes
        validation_datetimes = datetimes
        test_datetimes = datetimes
    elif method == SplitMethod.DAY:
        train_datetimes, validation_datetimes, test_datetimes = split_method(
            datetimes=datetimes,
            train_test_validation_split=train_test_validation_split,
            method="modulo",
        )
    elif method == SplitMethod.DAY_RANDOM:
        train_datetimes, validation_datetimes, test_datetimes = split_method(
            datetimes=datetimes,
            train_test_validation_split=train_test_validation_split,
            method="random",
            seed=seed,
        )
    elif method == SplitMethod.DAY_SPECIFIC:
        train_datetimes, validation_datetimes, test_datetimes = split_method(
            datetimes=datetimes,
            train_test_validation_split=train_test_validation_split,
            train_test_validation_specific=train_test_validation_specific,
            method="specific",
        )
    elif method == SplitMethod.WEEK:
        train_datetimes, validation_datetimes, test_datetimes = split_method(
            datetimes=datetimes,
            train_test_validation_split=train_test_validation_split,
            method="modulo",
            freq="W",
        )
    elif method == SplitMethod.WEEK_RANDOM:
        train_datetimes, validation_datetimes, test_datetimes = split_method(
            datetimes=datetimes,
            train_test_validation_split=train_test_validation_split,
            method="random",
            freq="W",
            seed=seed,
        )
    elif method == SplitMethod.YEAR_SPECIFIC:
        train_datetimes, validation_datetimes, test_datetimes = split_method(
            datetimes=datetimes,
            train_test_validation_split=train_test_validation_split,
            method="specific",
            freq="Y",
            seed=seed,
            train_test_validation_specific=train_test_validation_specific,
        )

    elif method == SplitMethod.DATE:
        train_datetimes, validation_datetimes, test_datetimes = split_by_dates(
            datetimes=datetimes,
            train_validation_datetime_split=train_validation_test_datetime_split[0],
            validation_test_datetime_split=train_validation_test_datetime_split[1],
        )

    elif method in [SplitMethod.DAY_RANDOM_TEST_YEAR, SplitMethod.DAY_RANDOM_TEST_DATE]:
        if method == SplitMethod.DAY_RANDOM_TEST_YEAR:
            # This method splits
            # 1. test set to be in one year, using 'train_test_validation_specific'
            # 2. train and validation by random day, using 'train_test_validation_split' on ratio
            #    how to split it.
            #
            # This allows us to create a test set for 2021, and train and validation for
            # random days not in 2021.

            # create test set
            train_datetimes, validation_datetimes, test_datetimes = split_method(
                datetimes=datetimes,
                train_test_validation_split=train_test_validation_split,
                method="specific",
                freq="Y",
                seed=seed,
                train_test_validation_specific=train_test_validation_specific,
            )
        elif method == SplitMethod.DAY_RANDOM_TEST_DATE:
            # This method splits
            # 1. test set from one date onwards
            # 2. train and validation by random day, using 'train_test_validation_split' on ratio
            #    how to split it.
            #
            # This allows us to create a test set from a specfic date e.g. 2020-07-01, and train
            # and validation for random days before that date.

            # create test set
            train_datetimes, validation_datetimes, test_datetimes = split_by_dates(
                datetimes=datetimes,
                train_validation_datetime_split=train_validation_test_datetime_split[0],
                validation_test_datetime_split=train_validation_test_datetime_split[1],
            )

        # join train and validation together, so they can then be split by random day.
        train_and_validation_datetimes = train_datetimes.append(validation_datetimes)

        # set split ratio to only be on train and validation
        train_validation_split = list(train_test_validation_split)
        train_validation_split[2] = 0
        train_validation_split = tuple(train_validation_split)

        # get train and validation methods
        train_datetimes, validation_datetimes, _ = split_method(
            datetimes=train_and_validation_datetimes,
            train_test_validation_split=train_validation_split,
            method="random",
            seed=seed,
        )

    else:
        raise ValueError(f"{method} for splitting day is not implemented")

    # Sanity check!
    if method != SplitMethod.SAME:
        assert len(train_datetimes.intersection(validation_datetimes)) == 0
        assert len(train_datetimes.intersection(test_datetimes)) == 0
        assert len(test_datetimes.intersection(validation_datetimes)) == 0

    assert train_datetimes.unique
    assert validation_datetimes.unique
    assert test_datetimes.unique

    split_datetimes = SplitDateTimes(
        train=train_datetimes, validation=validation_datetimes, test=test_datetimes
    )

    logger.debug("Split data done!")
    for split_name, dt in split_datetimes._asdict().items():
        if len(dt) == 0:
            # only a warning is made as this may happen during unittests
            logger.warning(f"{split_name} has {len(dt):,d} datetimes")
        else:
            logger.debug(f"{split_name} has {len(dt):,d} datetimes, from {dt[0]} to {dt[-1]}")

    return split_datetimes



"""Selection datapipes"""
from .drop_national_gsp import DropGSPIterDataPipe as DropGSP
from .location_picker import LocationPickerIterDataPipe as LocationPicker
from .offset_t0 import OffsetT0IterDataPipe as OffsetT0
from .select_id import SelectIDIterDataPipe as SelectID
from .select_live_t0_time import SelectLiveT0TimeIterDataPipe as SelectLiveT0Time
from .select_live_time_slice import SelectLiveTimeSliceIterDataPipe as SelectLiveTimeSlice
from .select_overlapping_time_slices import (
    SelectOverlappingTimeSliceIterDataPipe as SelectOverlappingTimeSlice,
)
from .select_spatial_slice import SelectSpatialSliceMetersIterDataPipe as SelectSpatialSliceMeters
from .select_spatial_slice import SelectSpatialSlicePixelsIterDataPipe as SelectSpatialSlicePixels
from .select_t0_time import SelectT0TimeIterDataPipe as SelectT0Time
from .select_time_periods import SelectTimePeriodsIterDataPipe as SelectTimePeriods
from .select_time_slice import SelectTimeSliceIterDataPipe as SelectTimeSlice
from .select_train_test import SelectTrainTestTimePeriodsIterDataPipe as SelectTrainTestTimePeriod
from .select_channels import SelectChannelsIterDataPipe as SelectChannels



"""Create the training/validation datapipe for training the national MetNet/-2 Model"""
import datetime
import logging
from datetime import timedelta
from pathlib import Path
from typing import Union

import xarray
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.config.model import Configuration
from ocf_datapipes.convert import ConvertGSPToNumpy
from ocf_datapipes.load import OpenConfiguration, OpenGSP, OpenNWP, OpenPVFromNetCDF, OpenSatellite
from ocf_datapipes.select import DropGSP, LocationPicker
from ocf_datapipes.transform.xarray import PreProcessMetNet
from ocf_datapipes.utils.consts import NWP_MEAN, NWP_STD, SAT_MEAN, SAT_MEAN_DA, SAT_STD, SAT_STD_DA

xarray.set_options(keep_attrs=True)
logger = logging.getLogger("metnet_datapipe")
logger.setLevel(logging.DEBUG)


def normalize_gsp(x):  # So it can be pickled
    """
    Normalize the GSP data

    Args:
        x: Input DataArray

    Returns:
        Normalized DataArray
    """
    return x / x.capacity_megawatt_power


def metnet_national_datapipe(
    configuration_filename: Union[Path, str],
    use_sun: bool = True,
    use_nwp: bool = True,
    use_sat: bool = True,
    use_hrv: bool = True,
    use_pv: bool = True,
    max_num_pv_systems: int = -1,
    start_time: datetime.datetime = datetime.datetime(2014, 1, 1),
    end_time: datetime.datetime = datetime.datetime(2023, 1, 1),
) -> IterDataPipe:
    """
    Make GSP national data pipe

    Currently only has GSP and NWP's in them

    Args:
        configuration_filename: the configruation filename for the pipe
        use_sun: Whether to add sun features or not
        use_pv: Whether to use PV input or not
        use_hrv: Whether to use HRV Satellite or not
        use_sat: Whether to use non-HRV Satellite or not
        use_nwp: Whether to use NWP or not
        max_num_pv_systems: max number of PV systems to include, <= 0 if no sampling
        start_time: Start time to select on
        end_time: End time to select from

    Returns: datapipe
    """

    # load configuration
    config_datapipe = OpenConfiguration(configuration_filename)
    configuration: Configuration = next(iter(config_datapipe))

    # Check which modalities to use
    if use_nwp:
        use_nwp = True if configuration.input_data.nwp.nwp_zarr_path != "" else False
    if use_pv:
        use_pv = True if configuration.input_data.pv.pv_files_groups[0].pv_filename != "" else False
    if use_sat:
        use_sat = True if configuration.input_data.satellite.satellite_zarr_path != "" else False
    if use_hrv:
        use_hrv = (
            True if configuration.input_data.hrvsatellite.hrvsatellite_zarr_path != "" else False
        )
    print(f"NWP: {use_nwp} Sat: {use_sat}, HRV: {use_hrv} PV: {use_pv} Sun: {use_sun}")
    # Load GSP national data
    logger.debug("Opening GSP Data")
    gsp_datapipe = OpenGSP(
        gsp_pv_power_zarr_path=configuration.input_data.gsp.gsp_zarr_path
    ).select_train_test_time(start_time, end_time)

    gsp_datapipe, gsp_loc_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0]).fork(2)

    location_datapipe = LocationPicker(gsp_loc_datapipe)

    logger.debug("Add t0 idx and normalize")

    gsp_datapipe, gsp_time_periods_datapipe, gsp_t0_datapipe = (
        gsp_datapipe.normalize(normalize_fn=normalize_gsp)
        .add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=30),
            history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        )
        .fork(3)
    )
    # get time periods
    # get contiguous time periods
    logger.debug("Getting contiguous time periods")
    gsp_time_periods_datapipe = gsp_time_periods_datapipe.get_contiguous_time_periods(
        sample_period_duration=timedelta(minutes=30),
        history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
    )

    secondary_datapipes = []

    # Load NWP data
    if use_nwp:
        logger.debug("Opening NWP Data")
        nwp_datapipe = OpenNWP(configuration.input_data.nwp.nwp_zarr_path).select_channels(
            configuration.input_data.nwp.nwp_channels
        )

        (
            nwp_datapipe,
            nwp_time_periods_datapipe,
        ) = nwp_datapipe.add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(hours=1),
            history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
        ).fork(
            2
        )

        nwp_time_periods_datapipe = nwp_time_periods_datapipe.get_contiguous_time_periods(
            sample_period_duration=timedelta(hours=3),  # Init times are 3 hours apart
            history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
            forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
            time_dim="init_time_utc",
        )
        secondary_datapipes.append(nwp_time_periods_datapipe)

    if use_sat:
        logger.debug("Opening Satellite Data")
        sat_datapipe = OpenSatellite(
            configuration.input_data.satellite.satellite_zarr_path
        ).select_channels(configuration.input_data.satellite.satellite_channels)
        (
            sat_datapipe,
            sat_time_periods_datapipe,
        ) = sat_datapipe.add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
        ).fork(
            2
        )

        sat_time_periods_datapipe = sat_time_periods_datapipe.get_contiguous_time_periods(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
            forecast_duration=timedelta(minutes=1),
        )
        secondary_datapipes.append(sat_time_periods_datapipe)

    if use_hrv:
        logger.debug("Opening HRV Satellite Data")
        sat_hrv_datapipe = OpenSatellite(
            configuration.input_data.hrvsatellite.hrvsatellite_zarr_path
        )

        (
            sat_hrv_datapipe,
            sat_hrv_time_periods_datapipe,
        ) = sat_hrv_datapipe.add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(
                minutes=configuration.input_data.hrvsatellite.history_minutes
            ),
        ).fork(
            2
        )
        sat_hrv_time_periods_datapipe = sat_hrv_time_periods_datapipe.get_contiguous_time_periods(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(
                minutes=configuration.input_data.hrvsatellite.history_minutes
            ),
            forecast_duration=timedelta(minutes=1),
        )
        secondary_datapipes.append(sat_hrv_time_periods_datapipe)

    if use_pv:
        logger.debug("Opening PV")
        pv_datapipe, pv_location_datapipe = OpenPVFromNetCDF(
            pv_power_filename=configuration.input_data.pv.pv_files_groups[0].pv_filename,
            pv_metadata_filename=configuration.input_data.pv.pv_files_groups[
                0
            ].pv_metadata_filename,
        ).fork(2)

        logger.debug("Add t0 idx")
        (
            pv_datapipe,
            pv_time_periods_datapipe,
        ) = pv_datapipe.add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        ).fork(
            2
        )

        pv_time_periods_datapipe = pv_time_periods_datapipe.get_contiguous_time_periods(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
            forecast_duration=timedelta(minutes=1),
        )
        secondary_datapipes.append(pv_time_periods_datapipe)

    # find joint overlapping timer periods
    logger.debug("Getting joint time periods")
    overlapping_datapipe = gsp_time_periods_datapipe.select_overlapping_time_slice(
        secondary_datapipes=secondary_datapipes,
    )

    (
        gsp_time_periods,
        nwp_time_periods,
        sat_time_periods,
        sat_hrv_time_periods,
        pv_time_periods,
    ) = overlapping_datapipe.fork(5)

    # select time periods
    gsp_t0_datapipe = gsp_t0_datapipe.select_time_periods(time_periods=gsp_time_periods)

    # select t0 periods
    logger.debug("Select t0 joint")
    (
        gsp_t0_datapipe,
        nwp_t0_datapipe,
        sat_t0_datapipe,
        sat_hrv_t0_datapipe,
        pv_t0_datapipe,
    ) = gsp_t0_datapipe.select_t0_time().fork(5)

    # take pv time slices
    logger.debug("Take GSP time slices")
    gsp_datapipe = gsp_datapipe.select_time_slice(
        t0_datapipe=gsp_t0_datapipe,
        history_duration=timedelta(minutes=0),
        forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
        sample_period_duration=timedelta(minutes=30),
    )

    if use_nwp:
        # take nwp time slices
        logger.debug("Take NWP time slices")
        nwp_datapipe = nwp_datapipe.convert_to_nwp_target_time(
            t0_datapipe=nwp_t0_datapipe,
            sample_period_duration=timedelta(hours=1),
            history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
            forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
        ).normalize(mean=NWP_MEAN, std=NWP_STD)

    if use_sat:
        logger.debug("Take Satellite time slices")
        # take sat time slices
        sat_datapipe = sat_datapipe.select_time_slice(
            t0_datapipe=sat_t0_datapipe,
            history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
            forecast_duration=timedelta(minutes=0),
            sample_period_duration=timedelta(minutes=5),
        ).normalize(mean=SAT_MEAN_DA, std=SAT_STD_DA)

    if use_hrv:
        logger.debug("Take HRV Satellite time slices")
        sat_hrv_datapipe = sat_hrv_datapipe.select_time_slice(
            t0_datapipe=sat_hrv_t0_datapipe,
            history_duration=timedelta(
                minutes=configuration.input_data.hrvsatellite.history_minutes
            ),
            forecast_duration=timedelta(minutes=0),
            sample_period_duration=timedelta(minutes=5),
        ).normalize(mean=SAT_MEAN["HRV"], std=SAT_STD["HRV"])

    if use_pv:
        logger.debug("Take PV Time Slices")
        # take pv time slices
        if use_sat:
            sat_datapipe, image_datapipe = sat_datapipe.fork(2)
        elif use_hrv:
            sat_hrv_datapipe, image_datapipe = sat_hrv_datapipe.fork(2)
        elif use_nwp:
            nwp_datapipe, image_datapipe = nwp_datapipe.fork(2)

        pv_datapipe = pv_datapipe.select_time_slice(
            t0_datapipe=pv_t0_datapipe,
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
            forecast_duration=timedelta(minutes=0),
            sample_period_duration=timedelta(minutes=5),
        ).create_pv_image(image_datapipe, normalize=True, max_num_pv_systems=max_num_pv_systems)

    # Now combine in the MetNet format
    modalities = []
    if use_nwp:
        modalities.append(nwp_datapipe)
    if use_hrv:
        modalities.append(sat_hrv_datapipe)
    if use_sat:
        modalities.append(sat_datapipe)
    if use_pv:
        modalities.append(pv_datapipe)

    combined_datapipe = PreProcessMetNet(
        modalities,
        location_datapipe=location_datapipe,
        center_width=500_000,
        center_height=1_000_000,
        context_height=10_000_000,
        context_width=10_000_000,
        output_width_pixels=256,
        output_height_pixels=256,
        add_sun_features=use_sun,
    )

    gsp_datapipe = ConvertGSPToNumpy(gsp_datapipe)
    return combined_datapipe.zip(gsp_datapipe)  # Makes (Inputs, Label) tuples



"""Loading datapipes from the raw data"""
from ocf_datapipes.load.gsp.database import OpenGSPFromDatabaseIterDataPipe as OpenGSPFromDatabase
from ocf_datapipes.load.gsp.gsp import OpenGSPIterDataPipe as OpenGSP
from ocf_datapipes.load.gsp.gsp_national import OpenGSPNationalIterDataPipe as OpenGSPNational
from ocf_datapipes.load.pv.live import OpenPVFromDBIterDataPipe as OpenPVFromDB
from ocf_datapipes.load.pv.pv import OpenPVFromNetCDFIterDataPipe as OpenPVFromNetCDF

from .configuration import OpenConfigurationIterDataPipe as OpenConfiguration
from .nwp.nwp import OpenNWPIterDataPipe as OpenNWP
from .nwp.nwp_id import OpenNWPIDIterDataPipe as OpenNWPID
from .satellite import OpenSatelliteIterDataPipe as OpenSatellite

try:
    import rioxarray  # Rioxarray is sometimes a pain to install, so only load this if its installed

    from .topographic import OpenTopographyIterDataPipe as OpenTopography
except ImportError:
    print("Rioxarray is not installed, so not importing OpenTopography")
    pass



import numpy as np

from ocf_datapipes.transform.xarray import PVFillNightNans


def test_pv_power_remove_data(passiv_datapipe):

    data_before = next(iter(passiv_datapipe))

    passiv_datapipe = PVFillNightNans(passiv_datapipe)
    data_after = next(iter(passiv_datapipe))

    assert np.isnan(data_before.values[-1, -1])
    assert not np.isnan(data_after.values[-1, -1])



"""Drop GSP output from xarray"""
from typing import List

from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("drop_gsp")
class DropGSPIterDataPipe(IterDataPipe):
    """Drops GSP"""

    def __init__(self, source_datapipe: IterDataPipe, gsps_to_keep: List[int] = slice(1, 317)):
        """
        Drop GSP National from the dataarray

        Args:
            source_datapipe: Datapipe emitting GSP xarray
            gsps_to_keep: List of GSP Id's to keep, by default all the 317 non-national ones
        """
        self.source_datapipe = source_datapipe
        self.gsps_to_keep = gsps_to_keep

    def __iter__(self):
        for xr_data in self.source_datapipe:
            # GSP ID 0 is national
            xr_data = xr_data.isel(gsp_id=self.gsps_to_keep)
            yield xr_data



"""Wrapper for Power Perceiver Production Data Pipeline"""
import logging
from datetime import timedelta
from pathlib import Path
from typing import Union

import xarray
from torchdata.datapipes.iter import IterDataPipe

import ocf_datapipes  # noqa
from ocf_datapipes.batch import MergeNumpyModalities
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import (
    OpenConfiguration,
    OpenGSPFromDatabase,
    OpenNWP,
    OpenPVFromDB,
    OpenSatellite,
    OpenTopography,
)
from ocf_datapipes.utils.consts import NWP_MEAN, NWP_STD, SAT_MEAN, SAT_STD, BatchKey

logger = logging.getLogger(__name__)
xarray.set_options(keep_attrs=True)


def power_perceiver_production_datapipe(configuration_filename: Union[Path, str]) -> IterDataPipe:
    """
    Create the Power Perceiver production pipeline using a configuration

    Args:
        configuration_filename: Name of the configuration

    Returns:
        DataPipe ready to be put in a Dataloader for production
    """
    ####################################
    #
    # Equivalent to PP's loading and filtering methods
    #
    #####################################
    # Normalize GSP and PV on whole dataset here
    config_datapipe = OpenConfiguration(configuration_filename)
    # TODO Pass the configuration through all the datapipes instead?
    configuration: Configuration = next(iter(config_datapipe))

    logger.debug("Opening Datasets")
    sat_hrv_datapipe = OpenSatellite(
        zarr_path=configuration.input_data.hrvsatellite.hrvsatellite_zarr_path
    )
    passiv_datapipe = OpenPVFromDB(
        providers=[pv_files.label for pv_files in configuration.input_data.pv.pv_files_groups],
        load_extra_minutes=configuration.input_data.pv.live_load_extra_minutes,
        history_minutes=configuration.input_data.pv.history_minutes,
    )

    nwp_datapipe = OpenNWP(configuration.input_data.nwp.nwp_zarr_path)
    topo_datapipe = OpenTopography(configuration.input_data.topographic.topographic_filename)
    gsp_datapipe, gso_loc_datapipe = (
        OpenGSPFromDatabase(
            history_minutes=configuration.input_data.gsp.history_minutes,
            interpolate_minutes=configuration.input_data.gsp.live_interpolate_minutes,
            load_extra_minutes=configuration.input_data.gsp.live_load_extra_minutes,
        )
        .drop_gsp()
        .fork(2)
    )
    logger.debug("Normalize GSP data")
    gsp_datapipe, gsp_t0_datapipe = (
        gsp_datapipe.normalize(normalize_fn=lambda x: x / x.capacity_megawatt_power)
        .add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=30),
            history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        )
        .fork(2)
    )
    logger.debug("Getting locations")
    (
        location_datapipe1,
        location_datapipe2,
        location_datapipe3,
        location_datapipe4,
    ) = gso_loc_datapipe.location_picker(return_all_locations=True).fork(4)

    logger.debug("Got locations")

    logger.debug("Making PV space slice")
    passiv_datapipe, pv_t0_datapipe = (
        passiv_datapipe.normalize(normalize_fn=lambda x: x / x.capacity_watt_power)
        .add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        )
        .select_spatial_slice_meters(
            location_datapipe=location_datapipe1,
            roi_width_meters=configuration.input_data.pv.pv_image_size_meters_width,
            roi_height_meters=configuration.input_data.pv.pv_image_size_meters_height,
        )
        .ensure_n_pv_systems_per_example(
            n_pv_systems_per_example=configuration.input_data.pv.n_pv_systems_per_example
        )
        .fork(2)
    )
    topo_datapipe = topo_datapipe.reproject_topography().normalize(
        calculate_mean_std_from_example=True
    )
    sat_hrv_datapipe, sat_t0_datapipe = (
        sat_hrv_datapipe.convert_satellite_to_int8()
        .add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(
                minutes=configuration.input_data.hrvsatellite.history_minutes
            ),
        )
        .select_spatial_slice_pixels(
            location_datapipe=location_datapipe2,
            roi_width_pixels=configuration.input_data.hrvsatellite.hrvsatellite_image_size_pixels_width,  # noqa
            roi_height_pixels=configuration.input_data.hrvsatellite.hrvsatellite_image_size_pixels_height,  # noqa
            y_dim_name="y_geostationary",
            x_dim_name="x_geostationary",
        )
        .fork(2)
    )

    logger.debug("Making NWP space slice")
    nwp_datapipe, nwp_t0_datapipe = (
        nwp_datapipe.add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(hours=1),
            history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
        )
        .select_spatial_slice_pixels(
            location_datapipe=location_datapipe3,
            roi_width_pixels=configuration.input_data.nwp.nwp_image_size_pixels_width
            * 16,  # TODO What to do here with configurations and such
            roi_height_pixels=configuration.input_data.nwp.nwp_image_size_pixels_height * 16,
            y_dim_name="y_osgb",
            x_dim_name="x_osgb",
        )
        .downsample(y_coarsen=16, x_coarsen=16)
        .fork(2)
    )

    nwp_t0_datapipe = nwp_t0_datapipe.select_live_t0_time(dim_name="init_time_utc")
    gsp_t0_datapipe = gsp_t0_datapipe.select_live_t0_time()
    sat_t0_datapipe = sat_t0_datapipe.select_live_t0_time()
    pv_t0_datapipe = pv_t0_datapipe.select_live_t0_time()

    logger.debug("Making GSP Time slices")
    gsp_datapipe = (
        gsp_datapipe.select_live_time_slice(
            t0_datapipe=gsp_t0_datapipe,
            history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        )
        .select_spatial_slice_meters(
            location_datapipe=location_datapipe4,
            roi_width_meters=10,
            roi_height_meters=10,
            dim_name="gsp_id",
        )
        .convert_gsp_to_numpy_batch()
        .extend_timesteps_to_future(
            forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
            sample_period_duration=timedelta(minutes=30),
        )
        .add_zeroed_future_data(key=BatchKey.gsp, time_key=BatchKey.gsp_time_utc)
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )
    logger.debug("Making Sat Time slices")
    sat_hrv_datapipe = (
        sat_hrv_datapipe.select_live_time_slice(
            t0_datapipe=sat_t0_datapipe,
            history_duration=timedelta(
                minutes=configuration.input_data.hrvsatellite.history_minutes
            ),
        )
        .normalize(mean=SAT_MEAN["HRV"] / 4, std=SAT_STD["HRV"] / 4)
        .map(
            lambda x: x.resample(time_utc="5T").interpolate("linear")
        )  # Interplate to 5 minutes incase its 15 minutes
        .convert_satellite_to_numpy_batch(is_hrv=True)
        .extend_timesteps_to_future(
            forecast_duration=timedelta(
                minutes=configuration.input_data.hrvsatellite.forecast_minutes
            ),
            sample_period_duration=timedelta(minutes=5),
        )
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )
    passiv_datapipe = (
        passiv_datapipe.select_live_time_slice(
            t0_datapipe=pv_t0_datapipe,
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        )
        .convert_pv_to_numpy_batch()
        .extend_timesteps_to_future(
            forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
            sample_period_duration=timedelta(minutes=5),
        )
        .add_zeroed_future_data(key=BatchKey.pv, time_key=BatchKey.pv_time_utc)
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )
    nwp_datapipe = (
        nwp_datapipe.convert_to_nwp_target_time(
            t0_datapipe=nwp_t0_datapipe,
            sample_period_duration=timedelta(hours=1),
            history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
            forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
        )
        .normalize(mean=NWP_MEAN, std=NWP_STD)
        .convert_nwp_to_numpy_batch()
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )

    ####################################
    #
    # Equivalent to PP's np_batch_processors
    #
    #####################################
    logger.debug("Combine all the data sources")
    combined_datapipe = (
        MergeNumpyModalities([gsp_datapipe, passiv_datapipe, sat_hrv_datapipe, nwp_datapipe])
        .align_gsp_to_5_min(batch_key_for_5_min_datetimes=BatchKey.hrvsatellite_time_utc)
        .encode_space_time()
        .save_t0_time()
        .add_sun_position(modality_name="hrvsatellite")
        .add_sun_position(modality_name="pv")
        .add_sun_position(modality_name="gsp")
        .add_sun_position(modality_name="gsp_5_min")
        .add_sun_position(modality_name="nwp_target_time")
        .add_topographic_data(topo_datapipe)
        .set_system_ids_to_one()
        .ensure_n_nwp_variables(num_variables=9)
        .change_float32()
    )

    return combined_datapipe



"""Selects time slice"""
from datetime import timedelta
from typing import Union

import numpy as np
import pandas as pd
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe, Zipper


@functional_datapipe("select_time_slice")
class SelectTimeSliceIterDataPipe(IterDataPipe):
    """Selects time slice"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        t0_datapipe: IterDataPipe,
        history_duration: timedelta,
        forecast_duration: timedelta,
        sample_period_duration: timedelta,
    ):
        """
        Selects time slice

        Args:
            source_datapipe: Datapipe of Xarray objects
            t0_datapipe: Datapipe of t0 times
            history_duration: History time used
            forecast_duration: Forecast time used
            sample_period_duration: Sample period of xarray data
        """
        self.source_datapipe = source_datapipe
        self.t0_datapipe = t0_datapipe
        self.history_duration = np.timedelta64(history_duration)
        self.forecast_duration = np.timedelta64(forecast_duration)
        self.sample_period_duration = sample_period_duration

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, t0 in Zipper(self.source_datapipe, self.t0_datapipe):
            t0_datetime_utc = pd.Timestamp(t0)
            start_dt = t0_datetime_utc - self.history_duration
            end_dt = t0_datetime_utc + self.forecast_duration
            xr_data = xr_data.sel(
                time_utc=slice(
                    start_dt.ceil(self.sample_period_duration),
                    end_dt.ceil(self.sample_period_duration),
                )
            )
            yield xr_data



"""NWP Loader"""
import logging
from pathlib import Path
from typing import Union

import pandas as pd
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

_log = logging.getLogger(__name__)


@functional_datapipe("open_nwp_id")
class OpenNWPIDIterDataPipe(IterDataPipe):
    """Opens NWP Zarr and yields it"""

    def __init__(self, netcdf_path: Union[Path, str]):
        """
        Opens NWP Zarr and yields it

        Args:
            netcdf_path: Path to the netcdf file
        """
        self.netcdf_path = netcdf_path

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Opens the NWP data"""
        _log.debug("Opening NWP data: %s", self.netcdf_path)
        ukv = open_nwp(self.netcdf_path)
        while True:
            yield ukv


def open_nwp(netcdf_path) -> xr.DataArray:
    """
    Opens the NWP data

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """
    _log.debug("Loading NWP")
    nwp = xr.load_dataset(netcdf_path, engine="h5netcdf", chunks="auto",)
    ukv: xr.DataArray = nwp["UKV"]
    del nwp
    ukv = ukv.transpose("init_time", "step", "variable", "id")
    ukv = ukv.rename({"init_time": "init_time_utc", "variable": "channel"})

    _log.debug("Resampling to 1 hour")
    ukv = ukv.resample(init_time_utc="1H").pad()

    # Sanity checks.
    time = pd.DatetimeIndex(ukv.init_time_utc)
    assert time.is_unique
    assert time.is_monotonic_increasing
    return ukv



from datetime import timedelta

from ocf_datapipes.select import SelectOverlappingTimeSlice, SelectTimePeriods, SelectTimeSlice
from ocf_datapipes.transform.xarray import GetContiguousT0TimePeriods


def test_time_picker(sat_hrv_datapipe, passiv_datapipe, gsp_datapipe):
    sat_time_datapipe = GetContiguousT0TimePeriods(
        sat_hrv_datapipe,
        sample_period_duration=timedelta(minutes=5),
        history_duration=timedelta(minutes=60),
        forecast_duration=timedelta(minutes=60),
    )
    pv_time_datapipe = GetContiguousT0TimePeriods(
        passiv_datapipe,
        sample_period_duration=timedelta(minutes=5),
        history_duration=timedelta(minutes=60),
        forecast_duration=timedelta(minutes=60),
    )
    gsp_time_datapipe = GetContiguousT0TimePeriods(
        gsp_datapipe,
        sample_period_duration=timedelta(minutes=5),
        history_duration=timedelta(minutes=60),
        forecast_duration=timedelta(minutes=60),
    )



"""Datapipe to check and optionally fill NaNs in data"""
from typing import Union

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("check_nans")
class CheckNaNsIterDataPipe(IterDataPipe):
    """Checks, and optionally fills, NaNs in Xarray Dataset"""

    def __init__(
        self, source_datapipe: IterDataPipe, dataset_name: str = None, fill_nans: bool = False
    ):
        """
        Checks and optionally fills NaNs in the data

        Args:
            source_datapipe: Datapipe emitting Xarray Datasets
            dataset_name: Optional name for dataset to check, if None, checks whole dataset
            fill_nans: Whether to fill NaNs with 0 or not
        """
        self.source_datapipe = source_datapipe
        self.dataset_name = dataset_name
        self.fill_nans = fill_nans

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """
        Checks for NaNs in data

        Returns:
            Original Xarray Dataset, after validation, and potentially with NaNs filled
        """
        for xr_data in self.source_datapipe:
            if self.fill_nans:
                if self.dataset_name is None:
                    xr_data = check_nan_and_fill_warning(data=xr_data)
                else:
                    xr_data[self.dataset_name] = check_nan_and_fill_warning(
                        data=xr_data[self.dataset_name]
                    )
            check_nan_and_inf(
                data=xr_data if self.dataset_name is None else xr_data[self.dataset_name]
            )
            yield xr_data


def check_nan_and_inf(data: xr.Dataset):
    """Check that all values are non NaNs and not infinite"""

    if np.isnan(data).any():
        message = "Some data values are NaNs. "

        # find out which example has nans in it
        for i in range(data.shape[0]):
            if np.isnan(data[i]).any():
                message += f" Nans in example {i}."
        raise Exception(message)

    if np.isinf(data).any():
        message = "Some data values are Infinite"
        raise Exception(message)


def check_nan_and_fill_warning(data: xr.Dataset) -> xr.Dataset:
    """Check that all values are non NaNs and not infinite"""

    if np.isnan(data).any():
        data = data.fillna(0)

    return data



"""Remove bad PV systems from the dataset"""
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("remove_bad_pv_systems")
class RemoveBadPVSystemsIterDataPipe(IterDataPipe):
    """Remove bad PV systems from the dataset"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Remove bad PV systems

        Args:
            source_datapipe: Datapipe of PV data
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> xr.DataArray:
        for xr_data in self.source_datapipe:
            yield xr_data



"""Ensure there is N PV systems per example"""
import logging

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("ensure_n_pv_systems_per_example")
class EnsureNPVSystemsPerExampleIterDataPipe(IterDataPipe):
    """Ensure there is N PV systems per example"""

    def __init__(self, source_datapipe: IterDataPipe, n_pv_systems_per_example: int, seed=None):
        """
        Ensure there is N PV systems per example

        Args:
            source_datapipe: Datapipe of PV data
            n_pv_systems_per_example: Number of PV systems to have in example
            seed: Random seed for choosing
        """
        self.source_datapipe = source_datapipe
        self.n_pv_systems_per_example = n_pv_systems_per_example
        self.rng = np.random.default_rng(seed=seed)

    def __iter__(self):
        for xr_data in self.source_datapipe:
            if len(xr_data.pv_system_id) > self.n_pv_systems_per_example:
                logger.debug(f"Reducing PV systems to  {self.n_pv_systems_per_example}")
                # More PV systems are available than we need. Reduce by randomly sampling:
                subset_of_pv_system_ids = self.rng.choice(
                    xr_data.pv_system_id, size=self.n_pv_systems_per_example, replace=False,
                )
                xr_data = xr_data.sel(pv_system_id=subset_of_pv_system_ids)
            elif len(xr_data.pv_system_id) < self.n_pv_systems_per_example:
                logger.debug("Padding out PV systems")
                # If we just used `choice(replace=True)` then there's a high chance
                # that the output won't include every available PV system but instead
                # will repeat some PV systems at the expense of leaving some on the table.
                # TODO: Don't repeat PV systems. Pad with NaNs and mask the loss. Issue #73.
                assert len(xr_data.pv_system_id) > 0, (
                    "There are no PV systems at all. " "We need at least one in an example"
                )
                n_random_pv_systems = self.n_pv_systems_per_example - len(xr_data.pv_system_id)
                allow_replacement = n_random_pv_systems > len(xr_data.pv_system_id)
                random_pv_system_ids = self.rng.choice(
                    xr_data.pv_system_id, size=n_random_pv_systems, replace=allow_replacement,
                )
                xr_data = xr.concat(
                    (xr_data, xr_data.sel(pv_system_id=random_pv_system_ids)), dim="pv_system_id",
                )
            yield xr_data



"""Configuration Loader"""
import logging

import fsspec
from pyaml_env import parse_config
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.config.model import Configuration

logger = logging.getLogger(__name__)


@functional_datapipe("open_config")
class OpenConfigurationIterDataPipe(IterDataPipe):
    """Open and return the configuration data"""

    def __init__(self, configuration_filename: str):
        """
        Open and return config data

        Args:
            configuration_filename: Filename to open
        """
        self.configuration_filename = configuration_filename

    def __iter__(self):
        """Open and return configuration file"""
        logger.debug(f"Going to open {self.configuration_filename}")
        with fsspec.open(self.configuration_filename, mode="r") as stream:
            configuration = parse_config(data=stream)

        logger.debug(f"Converting to Configuration ({configuration})")
        configuration = Configuration(**configuration)

        while True:
            yield configuration



"""Convert GSP to Numpy Array"""
import logging

import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("convert_gsp_to_numpy")
class ConvertGSPToNumpyIterDataPipe(IterDataPipe):
    """Convert GSP Xarray to Numpy Array"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert GSP Xarray to Numpy Array object

        Args:
            source_datapipe: Datapipe emitting GSP Xarray object
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> np.ndarray:
        """Convert from Xarray to Numpy array"""
        logger.debug("Converting GSP to numpy")
        for xr_data in self.source_datapipe:
            yield xr_data.values



""" Usual setup file for package """
# read the contents of your README file
from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
install_requires = (this_directory / "requirements.txt").read_text().splitlines()

setup(
    name="ocf_datapipes",
    version="0.5.1",
    license="MIT",
    description="Pytorch Datapipes built for use in Open Climate Fix's nowcasting/forecasting work",
    author="Jacob Bieker, Jack Kelly, Peter Dudfield",
    author_email="info@openclimatefix.org",
    company="Open Climate Fix Ltd",
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    extras_require={"all": ["cartopy==0.19.0.post1"]},
    include_package_data=True,
    packages=find_packages(),
)



""" Methods for splitting data into train, validation and test """
from typing import List, Tuple

import numpy as np
import pandas as pd

from ocf_datapipes.utils.split.model import (
    TrainValidationTestSpecific,
    default_train_test_validation_specific,
)


def split_method(
    datetimes: pd.DatetimeIndex,
    train_test_validation_split: Tuple[int] = (3, 1, 1),
    train_test_validation_specific: TrainValidationTestSpecific = (
        default_train_test_validation_specific
    ),
    method: str = "modulo",
    freq: str = "D",
    seed: int = 1234,
) -> (List[pd.Timestamp], List[pd.Timestamp], List[pd.Timestamp]):
    """
    Split the data into train, test and (optionally) validation sets.

    method: modulo
    If the split is (3,1,1) then, taking all the days in the dataset:
    - train data will have all days that are modulo 5 remainder 0,1 or 2,
    i.e 1st, 2nd, 3rd, 6th, 7th, 8th, 11th .... of the whole dataset
    - validation data will have all days that are modulo 5 remainder 3, i.e 4th, 9th, ...
    - test data will have all days that are modulo 5 remainder 4, i.e 5th, 10th ,  ...

    method: random
    If the split is (3,1,1) then
    - train data will have 60% of the data
    - validation data will have 20% of the data
    - test data will have have 20% of the data

    Args:
        datetimes: list of datetimes
        train_test_validation_split: how the split is made
        method: which method to use. Can be modulo or random
        freq: This can be D=day, W=week, M=month and Y=year. This means the data is divided up by
            different periods
        seed: random seed used to permutate the data for the 'random' method
        train_test_validation_specific: pydandic class of 'train', 'validation' and 'test'.
            These specify which data goes into which datasets

    Returns: train, validation and test datetimes

    """
    # find all the unique periods (dates, weeks, e.t.c)
    datetimes_period = pd.to_datetime(datetimes.to_period(freq).to_timestamp())
    unique_periods_in_dataset = datetimes_period.unique()

    # find total weights, and cumulative weights
    total_weights = sum(train_test_validation_split)
    cum_weights = np.cumsum(train_test_validation_split)

    if method == "modulo":
        # Method to split by module.
        # I.e 1st, 2nd, 3rd periods goes to train, 4th goes to validation, 5th goes to test and
        # repeat.

        # make which day indexes go i.e if the split is [3,1,1] then the
        # - train_ indexes = [0,1,2]
        # - validation_indexes = [3]
        # - test_indexes = [4]
        train_indexes = [i for i in range(total_weights) if i < cum_weights[0]]
        validation_indexes = [
            i for i in range(total_weights) if (i >= cum_weights[0]) & (i < cum_weights[1])
        ]
        test_indexes = [
            i for i in range(total_weights) if (i >= cum_weights[1]) & (i < cum_weights[2])
        ]

        # find all the unique periods (dates, weeks, e.t.c)
        unique_periods = pd.DataFrame(unique_periods_in_dataset, columns=["period"])
        unique_periods["modulo"] = unique_periods.index % total_weights

        train_periods = unique_periods[unique_periods["modulo"].isin(train_indexes)]["period"]
        validation_periods = unique_periods[unique_periods["modulo"].isin(validation_indexes)][
            "period"
        ]
        test_periods = unique_periods[unique_periods["modulo"].isin(test_indexes)]["period"]

    elif method == "random":

        # randomly sort indexes
        rng = np.random.default_rng(seed)
        unique_periods_in_dataset = rng.permutation(unique_periods_in_dataset)

        # find the train, validation, test indexes.
        #
        # For one year of data, for days, if the split is [3,1,1] for one year then the
        # - train_indexes - first 220 random dates
        # - validation_indexes - next 72 random dates
        # - test_indexes - next 72 random dates
        train_validation_split = int(
            cum_weights[0] / total_weights * len(unique_periods_in_dataset)
        )
        validation_test_split = int(cum_weights[1] / total_weights * len(unique_periods_in_dataset))

        train_periods = pd.to_datetime(unique_periods_in_dataset[0:train_validation_split])
        validation_periods = pd.to_datetime(
            unique_periods_in_dataset[train_validation_split:validation_test_split]
        )
        test_periods = pd.to_datetime(unique_periods_in_dataset[validation_test_split:])

    elif method == "specific":

        train_periods = unique_periods_in_dataset[
            unique_periods_in_dataset.isin(train_test_validation_specific.train)
        ]
        validation_periods = unique_periods_in_dataset[
            unique_periods_in_dataset.isin(train_test_validation_specific.validation)
        ]
        test_periods = unique_periods_in_dataset[
            unique_periods_in_dataset.isin(train_test_validation_specific.test)
        ]

    else:
        raise Exception(f'method ({method}) must be in ["random", "modulo"]')

    train = datetimes[datetimes_period.isin(train_periods)]
    validation = datetimes[datetimes_period.isin(validation_periods)]
    test = datetimes[datetimes_period.isin(test_periods)]

    return train, validation, test


def split_by_dates(
    datetimes: pd.DatetimeIndex,
    train_validation_datetime_split: pd.Timestamp,
    validation_test_datetime_split: pd.Timestamp,
) -> (List[pd.Timestamp], List[pd.Timestamp], List[pd.Timestamp]):
    """
    Split datetimes into train, validation and test by two specific datetime splits

    Note that the 'train_validation_datetime_split' should be less than the
    'validation_test_datetime_split'

    Args:
        datetimes: list of datetimes
        train_validation_datetime_split: the datetime which will split the train and validation
            datetimes.
        For example if this is '2021-01-01' then the train datetimes will end by '2021-01-01' and
            the validation datetimes will start at '2021-01-01'.
        validation_test_datetime_split: the datetime which will split the validation and
            test datetimes

    Returns: train, validation and test datetimes

    """
    assert train_validation_datetime_split <= validation_test_datetime_split

    train = datetimes[datetimes < train_validation_datetime_split]
    validation = datetimes[
        (datetimes >= train_validation_datetime_split)
        & (datetimes < validation_test_datetime_split)
    ]
    test = datetimes[datetimes >= validation_test_datetime_split]

    return train, validation, test



""" Configuration model for the dataset.

All paths must include the protocol prefix.  For local files,
it's sufficient to just start with a '/'.  For aws, start with 's3://',
for gcp start with 'gs://'.

This file is mostly about _configuring_ the DataSources.

Separate Pydantic models in
`nowcasting_dataset/data_sources/<data_source_name>/<data_source_name>_model.py`
are used to validate the values of the data itself.

"""
import logging
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union

import git
import numpy as np
import pandas as pd
from nowcasting_datamodel.models.pv import providers, pv_output, solar_sheffield_passiv
from pathy import Pathy
from pydantic import BaseModel, Field, root_validator, validator

# nowcasting_dataset imports
from ocf_datapipes.utils.consts import (
    DEFAULT_N_GSP_PER_EXAMPLE,
    DEFAULT_N_PV_SYSTEMS_PER_EXAMPLE,
    NWP_VARIABLE_NAMES,
    SAT_VARIABLE_NAMES,
)
from ocf_datapipes.utils.split import split

IMAGE_SIZE_PIXELS = 64
IMAGE_SIZE_PIXELS_FIELD = Field(
    IMAGE_SIZE_PIXELS, description="The number of pixels of the region of interest."
)
METERS_PER_PIXEL_FIELD = Field(2000, description="The number of meters per pixel.")
METERS_PER_ROI = Field(128_000, description="The number of meters of region of interest.")

logger = logging.getLogger(__name__)


class Base(BaseModel):
    """Pydantic Base model where no extras can be added"""

    class Config:
        """config class"""

        extra = "forbid"  # forbid use of extra kwargs


class General(Base):
    """General pydantic model"""

    name: str = Field("example", description="The name of this configuration file.")
    description: str = Field(
        "example configuration", description="Description of this configuration file"
    )


class Git(Base):
    """Git model"""

    hash: str = Field(
        ..., description="The git hash of nowcasting_dataset when a dataset is created."
    )
    message: str = Field(..., description="The git message for when a dataset is created.")
    committed_date: datetime = Field(
        ..., description="The git datestamp for when a dataset is created."
    )


class DataSourceMixin(Base):
    """Mixin class, to add forecast and history minutes"""

    forecast_minutes: int = Field(
        None,
        ge=0,
        description="how many minutes to forecast in the future. "
        "If set to None, the value is defaulted to InputData.default_forecast_minutes",
    )
    history_minutes: int = Field(
        None,
        ge=0,
        description="how many historic minutes to use. "
        "If set to None, the value is defaulted to InputData.default_history_minutes",
    )

    log_level: str = Field(
        "DEBUG",
        description="The logging level for this data source. T"
        "his is the default value and can be set in each data source",
    )

    @property
    def seq_length_30_minutes(self):
        """How many steps are there in 30 minute datasets"""
        return int(np.ceil((self.history_minutes + self.forecast_minutes) / 30 + 1))

    @property
    def seq_length_5_minutes(self):
        """How many steps are there in 5 minute datasets"""
        return int(np.ceil((self.history_minutes + self.forecast_minutes) / 5 + 1))

    @property
    def seq_length_60_minutes(self):
        """How many steps are there in 60 minute datasets"""
        return int(np.ceil((self.history_minutes + self.forecast_minutes) / 60 + 1))

    @property
    def history_seq_length_5_minutes(self):
        """How many historical steps are there in 5 minute datasets"""
        return int(np.ceil(self.history_minutes / 5))

    @property
    def history_seq_length_30_minutes(self):
        """How many historical steps are there in 30 minute datasets"""
        return int(np.ceil(self.history_minutes / 30))

    @property
    def history_seq_length_60_minutes(self):
        """How many historical steps are there in 60 minute datasets"""
        return int(np.ceil(self.history_minutes / 60))


class TimeResolutionMixin(Base):
    """Time resolution mix in"""

    # TODO: Issue #584: Rename to `sample_period_minutes`
    time_resolution_minutes: int = Field(
        5,
        description="The temporal resolution (in minutes) of the satellite images."
        "Note that this needs to be divisible by 5.",
    )

    @validator("time_resolution_minutes")
    def forecast_minutes_divide_by_5(cls, v):
        """Validate 'forecast_minutes'"""
        assert v % 5 == 0, f"The time resolution ({v}) is not divisible by 5"
        return v


class StartEndDatetimeMixin(Base):
    """Mixin class to add start and end date"""

    start_datetime: datetime = Field(
        datetime(2020, 1, 1),
        description="Load date from data sources from this date. "
        "If None, this will get overwritten by InputData.start_date. ",
    )
    end_datetime: datetime = Field(
        datetime(2021, 9, 1),
        description="Load date from data sources up to this date. "
        "If None, this will get overwritten by InputData.start_date. ",
    )

    @root_validator
    def check_start_and_end_datetime(cls, values):
        """
        Make sure start datetime is before end datetime
        """

        start_datetime = values["start_datetime"]
        end_datetime = values["end_datetime"]

        # check start datetime is less than end datetime
        if start_datetime >= end_datetime:
            message = (
                f"Start datetime ({start_datetime}) "
                f"should be less than end datetime ({end_datetime})"
            )
            logger.error(message)
            assert Exception(message)

        return values


class PVFiles(BaseModel):
    """Model to hold pv file and metadata file"""

    pv_filename: str = Field(
        "gs://solar-pv-nowcasting-data/PV/PVOutput.org/UK_PV_timeseries_batch.nc",
        description="The NetCDF files holding the solar PV power timeseries.",
    )
    pv_metadata_filename: str = Field(
        "gs://solar-pv-nowcasting-data/PV/PVOutput.org/UK_PV_metadata.csv",
        description="Tthe CSV files describing each PV system.",
    )

    label: str = Field(pv_output, description="Label of where the pv data came from")

    @validator("label")
    def v_label0(cls, v):
        """Validate 'label'"""
        if v not in providers:
            message = f"provider {v} not in {providers}"
            logger.error(message)
            raise Exception(message)
        return v


class PV(DataSourceMixin, StartEndDatetimeMixin):
    """PV configuration model"""

    pv_files_groups: List[PVFiles] = [PVFiles()]

    n_pv_systems_per_example: int = Field(
        DEFAULT_N_PV_SYSTEMS_PER_EXAMPLE,
        description="The number of PV systems samples per example. "
        "If there are less in the ROI then the data is padded with zeros. ",
    )
    pv_image_size_meters_height: int = METERS_PER_ROI
    pv_image_size_meters_width: int = METERS_PER_ROI
    get_center: bool = Field(
        False,
        description="If the batches are centered on one PV system (or not). "
        "The other options is to have one GSP at the center of a batch. "
        "Typically, get_center would be set to true if and only if "
        "PVDataSource is used to define the geospatial positions of each example.",
    )

    pv_filename: str = Field(
        None, description="The NetCDF files holding the solar PV power timeseries.",
    )
    pv_metadata_filename: str = Field(
        None, description="Tthe CSV files describing each PV system.",
    )

    is_live: bool = Field(
        False, description="Option if to use live data from the nowcasting pv database"
    )

    live_interpolate_minutes: int = Field(
        30, description="The number of minutes we allow PV data to interpolate"
    )
    live_load_extra_minutes: int = Field(
        0,
        description="The number of extra minutes in the past we should load. Then the recent "
        "values can be interpolated, and the extra minutes removed. This is "
        "because some live data takes ~1 hour to come in.",
    )

    @classmethod
    def model_validation(cls, v):
        """Move old way of storing filenames to new way"""

        if (v.pv_filename is not None) and (v.pv_metadata_filename is not None):
            logger.warning(
                "Loading pv files the old way, and moving them the new way. "
                "Please update configuration file"
            )
            label = pv_output if "pvoutput" in v.pv_filename.lower() else solar_sheffield_passiv
            pv_file = PVFiles(
                pv_filename=v.pv_filename, pv_metadata_filename=v.pv_metadata_filename, label=label
            )
            v.pv_files_groups = [pv_file]
            v.pv_filename = None
            v.pv_metadata_filename = None

        return v


class Satellite(DataSourceMixin, TimeResolutionMixin):
    """Satellite configuration model"""

    satellite_zarr_path: str = Field(
        "gs://solar-pv-nowcasting-data/satellite/EUMETSAT/SEVIRI_RSS/OSGB36/all_zarr_int16_single_timestep.zarr",  # noqa: E501
        description="The path which holds the satellite zarr.",
    )
    satellite_channels: tuple = Field(
        SAT_VARIABLE_NAMES[1:], description="the satellite channels that are used"
    )
    satellite_image_size_pixels_height: int = Field(
        IMAGE_SIZE_PIXELS_FIELD.default // 3,
        description="The number of pixels of the height of the region of interest"
        " for non-HRV satellite channels.",
    )
    satellite_image_size_pixels_width: int = Field(
        IMAGE_SIZE_PIXELS_FIELD.default // 3,
        description="The number of pixels of the width of the region "
        "of interest for non-HRV satellite channels.",
    )
    satellite_meters_per_pixel: int = Field(
        METERS_PER_PIXEL_FIELD.default * 3,
        description="The number of meters per pixel for non-HRV satellite channels.",
    )

    keep_dawn_dusk_hours: int = Field(
        0,
        description="The number hours around dawn and dusk that should be keep. "
        "I.e 'keep_dawn_dusk_hours'=2,"
        " then if dawn if 07.00, "
        " then data is keep from 06.00",
    )

    is_live: bool = Field(
        False,
        description="Option if to use live data from the satelite consumer. "
        "This is useful becasuse the data is about ~30 mins behind, "
        "so we need to expect that",
    )

    live_delay_minutes: int = Field(
        30, description="The expected delay in minutes of the satellite data"
    )


class HRVSatellite(DataSourceMixin, TimeResolutionMixin):
    """Satellite configuration model for HRV data"""

    hrvsatellite_zarr_path: str = Field(
        "gs://solar-pv-nowcasting-data/satellite/EUMETSAT/SEVIRI_RSS/OSGB36/all_zarr_int16_single_timestep.zarr",  # noqa: E501
        description="The path which holds the satellite zarr.",
    )

    hrvsatellite_channels: tuple = Field(
        SAT_VARIABLE_NAMES[0:1], description="the satellite channels that are used"
    )
    # HRV is 3x the resolution, so to cover the same area, its 1/3 the meters per pixel and 3
    # time the number of pixels
    hrvsatellite_image_size_pixels_height: int = IMAGE_SIZE_PIXELS_FIELD
    hrvsatellite_image_size_pixels_width: int = IMAGE_SIZE_PIXELS_FIELD
    hrvsatellite_meters_per_pixel: int = METERS_PER_PIXEL_FIELD

    keep_dawn_dusk_hours: int = Field(
        0,
        description="The number hours around dawn and dusk that should be keep. "
        "I.e 'keep_dawn_dusk_hours'=2,"
        " then if dawn if 07.00, "
        " then data is keep from 06.00",
    )

    is_live: bool = Field(
        False,
        description="Option if to use live data from the satelite consumer. "
        "This is useful becasuse the data is about ~30 mins behind, "
        "so we need to expect that",
    )

    live_delay_minutes: int = Field(
        30, description="The expected delay in minutes of the satellite data"
    )


class OpticalFlow(DataSourceMixin, TimeResolutionMixin):
    """Optical Flow configuration model"""

    opticalflow_zarr_path: str = Field(
        "",
        description=(
            "The satellite Zarr data to use. If in doubt, use the same value as"
            " satellite.satellite_zarr_path."
        ),
    )

    # history_minutes, set in DataSourceMixin.
    # Duration of historical data to use when computing the optical flow field.
    # For example, set to 5 to use just two images: the t-1 and t0 images.  Set to 10 to
    # compute the optical flow field separately for the image pairs (t-2, t-1), and
    # (t-1, t0) and to use the mean optical flow field.

    # forecast_minutes, set in DataSourceMixin.
    # Duration of the optical flow predictions.

    opticalflow_meters_per_pixel: int = METERS_PER_PIXEL_FIELD
    opticalflow_input_image_size_pixels_height: int = Field(
        IMAGE_SIZE_PIXELS * 2,
        description=(
            "The *input* image height (i.e. the image size to load off disk)."
            " This should be larger than output_image_size_pixels to provide sufficient border to"
            " mean that, even after the image has been flowed, all edges of the output image are"
            " real pixels values, and not NaNs."
        ),
    )
    opticalflow_output_image_size_pixels_height: int = Field(
        IMAGE_SIZE_PIXELS,
        description=(
            "The height of the images after optical flow has been applied. The output image is a"
            " center-crop of the input image, after it has been flowed."
        ),
    )
    opticalflow_input_image_size_pixels_width: int = Field(
        IMAGE_SIZE_PIXELS * 2,
        description=(
            "The *input* image width (i.e. the image size to load off disk)."
            " This should be larger than output_image_size_pixels to provide sufficient border to"
            " mean that, even after the image has been flowed, all edges of the output image are"
            " real pixels values, and not NaNs."
        ),
    )
    opticalflow_output_image_size_pixels_width: int = Field(
        IMAGE_SIZE_PIXELS,
        description=(
            "The width of the images after optical flow has been applied. The output image is a"
            " center-crop of the input image, after it has been flowed."
        ),
    )
    opticalflow_channels: tuple = Field(
        SAT_VARIABLE_NAMES[1:], description="the satellite channels that are used"
    )
    opticalflow_source_data_source_class_name: str = Field(
        "SatelliteDataSource",
        description=(
            "Either SatelliteDataSource or HRVSatelliteDataSource."
            "  The name of the DataSource that will load the satellite images."
        ),
    )


class NWP(DataSourceMixin):
    """NWP configuration model"""

    # TODO change to nwp_path, as it could be a netcdf now.
    # https://github.com/openclimatefix/nowcasting_dataset/issues/582
    nwp_zarr_path: str = Field(
        "gs://solar-pv-nowcasting-data/NWP/UK_Met_Office/UKV__2018-01_to_2019-12__chunks__variable10__init_time1__step1__x548__y704__.zarr",  # noqa: E501
        description="The path which holds the NWP zarr.",
    )
    nwp_channels: tuple = Field(NWP_VARIABLE_NAMES, description="the channels used in the nwp data")
    nwp_image_size_pixels_height: int = IMAGE_SIZE_PIXELS_FIELD
    nwp_image_size_pixels_width: int = IMAGE_SIZE_PIXELS_FIELD
    nwp_meters_per_pixel: int = METERS_PER_PIXEL_FIELD


class GSP(DataSourceMixin, StartEndDatetimeMixin):
    """GSP configuration model"""

    gsp_zarr_path: str = Field("gs://solar-pv-nowcasting-data/PV/GSP/v2/pv_gsp.zarr")
    n_gsp_per_example: int = Field(
        DEFAULT_N_GSP_PER_EXAMPLE,
        description="The number of GSP samples per example. "
        "If there are less in the ROI then the data is padded with zeros. ",
    )
    gsp_image_size_pixels_height: int = IMAGE_SIZE_PIXELS_FIELD
    gsp_image_size_pixels_width: int = IMAGE_SIZE_PIXELS_FIELD
    gsp_meters_per_pixel: int = METERS_PER_PIXEL_FIELD
    metadata_only: bool = Field(False, description="Option to only load metadata.")

    is_live: bool = Field(
        False, description="Option if to use live data from the nowcasting GSP/Forecast database"
    )

    live_interpolate_minutes: int = Field(
        60, description="The number of minutes we allow GSP data to be interpolated"
    )
    live_load_extra_minutes: int = Field(
        60,
        description="The number of extra minutes in the past we should load. Then the recent "
        "values can be interpolated, and the extra minutes removed. This is "
        "because some live data takes ~1 hour to come in.",
    )

    @validator("history_minutes")
    def history_minutes_divide_by_30(cls, v):
        """Validate 'history_minutes'"""
        assert v % 30 == 0  # this means it also divides by 5
        return v

    @validator("forecast_minutes")
    def forecast_minutes_divide_by_30(cls, v):
        """Validate 'forecast_minutes'"""
        assert v % 30 == 0  # this means it also divides by 5
        return v


class Topographic(DataSourceMixin):
    """Topographic configuration model"""

    topographic_filename: str = Field(
        "gs://solar-pv-nowcasting-data/Topographic/europe_dem_1km_osgb.tif",
        description="Path to the GeoTIFF Topographic data source",
    )
    topographic_image_size_pixels_height: int = IMAGE_SIZE_PIXELS_FIELD
    topographic_image_size_pixels_width: int = IMAGE_SIZE_PIXELS_FIELD
    topographic_meters_per_pixel: int = METERS_PER_PIXEL_FIELD


class Sun(DataSourceMixin):
    """Sun configuration model"""

    sun_zarr_path: str = Field(
        "gs://solar-pv-nowcasting-data/Sun/v1/sun.zarr/",
        description="Path to the Sun data source i.e Azimuth and Elevation",
    )
    load_live: bool = Field(
        False, description="Option to load sun data on the fly, rather than from file"
    )

    elevation_limit: int = Field(
        10,
        description="The limit to the elevations for examples. "
        "Datetimes below this limits will be ignored",
    )


class InputData(Base):
    """
    Input data model.
    """

    pv: Optional[PV] = None
    satellite: Optional[Satellite] = None
    hrvsatellite: Optional[HRVSatellite] = None
    opticalflow: Optional[OpticalFlow] = None
    nwp: Optional[NWP] = None
    gsp: Optional[GSP] = None
    topographic: Optional[Topographic] = None
    sun: Optional[Sun] = None

    default_forecast_minutes: int = Field(
        60,
        ge=0,
        description="how many minutes to forecast in the future. "
        "This sets the default for all the data sources if they are not set.",
    )
    default_history_minutes: int = Field(
        30,
        ge=0,
        description="how many historic minutes are used. "
        "This sets the default for all the data sources if they are not set.",
    )
    data_source_which_defines_geospatial_locations: str = Field(
        "gsp",
        description=(
            "The name of the DataSource which will define the geospatial position of each example."
        ),
    )

    @property
    def default_seq_length_5_minutes(self):
        """How many steps are there in 5 minute datasets"""
        return int((self.default_history_minutes + self.default_forecast_minutes) / 5 + 1)

    @root_validator
    def set_forecast_and_history_minutes(cls, values):
        """
        Set default history and forecast values, if needed.

        Run through the different data sources and  if the forecast or history minutes are not set,
        then set them to the default values
        """
        # It would be much better to use nowcasting_dataset.data_sources.ALL_DATA_SOURCE_NAMES,
        # but that causes a circular import.
        ALL_DATA_SOURCE_NAMES = (
            "pv",
            "hrvsatellite",
            "satellite",
            "nwp",
            "gsp",
            "topographic",
            "sun",
            "opticalflow",
        )
        enabled_data_sources = [
            data_source_name
            for data_source_name in ALL_DATA_SOURCE_NAMES
            if values[data_source_name] is not None
        ]

        for data_source_name in enabled_data_sources:
            if values[data_source_name].forecast_minutes is None:
                values[data_source_name].forecast_minutes = values["default_forecast_minutes"]

            if values[data_source_name].history_minutes is None:
                values[data_source_name].history_minutes = values["default_history_minutes"]

        return values

    @classmethod
    def set_all_to_defaults(cls):
        """Returns an InputData instance with all fields set to their default values.

        Used for unittests.
        """
        return cls(
            pv=PV(),
            satellite=Satellite(),
            hrvsatellite=HRVSatellite(),
            nwp=NWP(),
            gsp=GSP(),
            topographic=Topographic(),
            sun=Sun(),
            opticalflow=OpticalFlow(),
        )


class OutputData(Base):
    """Output data model"""

    filepath: Union[str, Pathy] = Field(
        Pathy("gs://solar-pv-nowcasting-data/prepared_ML_training_data/v7/"),
        description=(
            "Where the data is saved to.  If this is running on the cloud then should include"
            " 'gs://' or 's3://'"
        ),
    )

    @validator("filepath")
    def filepath_pathy(cls, v):
        """Make sure filepath is a Pathy object"""
        return Pathy(v)


class Process(Base):
    """Pydantic model of how the data is processed"""

    seed: int = Field(1234, description="Random seed, so experiments can be repeatable")
    batch_size: int = Field(32, description="The number of examples per batch")
    t0_datetime_frequency: pd.Timedelta = Field(
        pd.Timedelta("5 minutes"),
        description=(
            "The temporal frequency at which t0 datetimes will be sampled."
            "  Can be any string that `pandas.Timedelta()` understands."
            "  For example, if this is set to '5 minutes', then, for each example, the t0 datetime"
            " could be at 0, 5, ..., 55 minutes past the hour.  If there are DataSources with a"
            " lower sample rate (e.g. half-hourly) then these lower-sample-rate DataSources will"
            " still produce valid examples.  For example, if a half-hourly DataSource is asked for"
            " an example with t0=12:05, history_minutes=60, forecast_minutes=60, then it will"
            " return data at 11:30, 12:00, 12:30, and 13:00."
        ),
    )
    split_method: split.SplitMethod = Field(
        split.SplitMethod.DAY_RANDOM_TEST_DATE,
        description=(
            "The method used to split the t0 datetimes into train, validation and test sets."
            " If the split method produces no t0 datetimes for any split_name, then"
            " n_<split_name>_batches must also be set to 0."
        ),
    )

    train_test_validation_split: List[int] = Field(
        [10, 1, 1],
        description=(
            "The ratio of how the entire dataset is split. "
            "Note different split methods interact different with these numbers"
        ),
    )

    n_train_batches: int = Field(
        250,
        description=(
            "Number of train batches.  Must be 0 if split_method produces no t0 datetimes for"
            " the train split"
        ),
    )
    n_validation_batches: int = Field(
        0,  # Currently not using any validation batches!
        description=(
            "Number of validation batches.  Must be 0 if split_method produces no t0 datetimes for"
            " the validation split"
        ),
    )
    n_test_batches: int = Field(
        10,
        description=(
            "Number of test batches.  Must be 0 if split_method produces no t0 datetimes for"
            " the test split."
        ),
    )
    upload_every_n_batches: int = Field(
        16,
        description=(
            "How frequently to move batches from the local temporary directory to the cloud bucket."
            "  If 0 then write batches directly to output_data.filepath, not to a temp directory."
        ),
    )

    local_temp_path: Path = Field(
        Path("~/temp/").expanduser(),
        description=(
            "This is only necessary if using a VM on a public cloud and when the finished batches"
            " will be uploaded to a cloud bucket. This is the local temporary path on the VM."
            "  This will be emptied."
        ),
    )

    @validator("local_temp_path")
    def local_temp_path_to_path_object_expanduser(cls, v):
        """
        Convert the local path to Path

        Convert the path in string format to a `pathlib.PosixPath` object
        and call `expanduser` on the latter.
        """
        return Path(v).expanduser()


class Configuration(Base):
    """Configuration model for the dataset"""

    general: General = General()
    input_data: InputData = InputData()
    output_data: OutputData = OutputData()
    process: Process = Process()
    git: Optional[Git] = None

    def set_base_path(self, base_path: str):
        """Append base_path to all paths. Mostly used for testing."""
        base_path = Pathy(base_path)
        path_attrs = [
            "pv.pv_filename",
            "pv.pv_metadata_filename",
            "satellite.satellite_zarr_path",
            "hrvsatellite.hrvsatellite_zarr_path",
            "nwp.nwp_zarr_path",
            "gsp.gsp_zarr_path",
        ]
        for cls_and_attr_name in path_attrs:
            cls_name, attribute = cls_and_attr_name.split(".")
            cls = getattr(self.input_data, cls_name)
            path = getattr(getattr(self.input_data, cls_name), attribute)
            path = base_path / path
            setattr(cls, attribute, path)
            setattr(self.input_data, cls_name, cls)


def set_git_commit(configuration: Configuration):
    """
    Set the git information in the configuration file

    Args:
        configuration: configuration object

    Returns: configuration object with git information

    """

    repo = git.Repo(search_parent_directories=True)
    git.refresh("/usr/bin/git")

    git_details = Git(
        hash=repo.head.object.hexsha,
        committed_date=datetime.fromtimestamp(repo.head.object.committed_date),
        message=repo.head.object.message,
    )

    configuration.git = git_details

    return configuration



"""Preprocessing for MetNet-type inputs"""
from typing import List

import numpy as np
import pandas as pd
import pvlib
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe, Zipper

from ocf_datapipes.utils.geospatial import (
    load_geostationary_area_definition_and_transform_osgb,
    load_geostationary_area_definition_and_transform_to_latlon,
    osgb_to_lat_lon,
)

ELEVATION_MEAN = 37.4
ELEVATION_STD = 12.7
AZIMUTH_MEAN = 177.7
AZIMUTH_STD = 41.7


@functional_datapipe("preprocess_metnet")
class PreProcessMetNetIterDataPipe(IterDataPipe):
    """Preprocess set of Xarray datasets similar to MetNet-1"""

    def __init__(
        self,
        source_datapipes: List[IterDataPipe],
        location_datapipe: IterDataPipe,
        context_width: int,
        context_height: int,
        center_width: int,
        center_height: int,
        output_height_pixels: int,
        output_width_pixels: int,
        add_sun_features: bool = False,
    ):
        """

        Processes set of Xarray datasets similar to MetNet

        In terms of taking all available source datapipes:
        1. selecting the same context area of interest
        2. Creating a center crop of the center_height, center_width
        3. Downsampling the context area of interest to the same shape as the center crop
        4. Stacking those context images on the center crop.
        5. Add Month, Day, Hour channels for each input time
        6. Add Sun position as well?

        This would be designed originally for NWP+Satellite+Topographic data sources.
        To add the PV power for lots of sites, the PV power would
        need to be able to be on a grid for the context/center
        crops and then for the downsample

        This also appends Lat/Lon coordinates to the stack,
         and returns a new Numpy array with the stacked data

        TODO Could also add the national PV as a set of Layers, so also GSP input

        Args:
            source_datapipes: Datapipes that emit xarray datasets
                with latitude/longitude coordinates included
            location_datapipe: Datapipe emitting location coordinate for center of example
            context_width: Width of the context area
            context_height: Height of the context area
            center_width: Center width of the area of interest
            center_height: Center height of the area of interest
            output_height_pixels: Output height in pixels
            output_width_pixels: Output width in pixels
            add_sun_features: Whether to calculate and
            add Sun elevation and azimuth for each center pixel
        """
        self.source_datapipes = source_datapipes
        self.location_datapipe = location_datapipe
        self.context_width = context_width
        self.context_height = context_height
        self.center_width = center_width
        self.center_height = center_height
        self.output_height_pixels = output_height_pixels
        self.output_width_pixels = output_width_pixels
        self.add_sun_features = add_sun_features

    def __iter__(self) -> np.ndarray:
        for xr_datas, location in Zipper(Zipper(*self.source_datapipes), self.location_datapipe):
            # TODO Use the Lat/Long coordinates of the center array for the lat/lon stuff
            centers = []
            contexts = []
            for xr_index, xr_data in enumerate(xr_datas):
                xr_context: xr.Dataset = _get_spatial_crop(
                    xr_data,
                    location=location,
                    roi_width_meters=self.context_width,
                    roi_height_meters=self.context_height,
                )

                xr_center: xr.Dataset = _get_spatial_crop(
                    xr_data,
                    location=location,
                    roi_width_meters=self.center_width,
                    roi_height_meters=self.center_height,
                )
                # Resamples to the same number of pixels for both center and contexts
                xr_center = _resample_to_pixel_size(
                    xr_center, self.output_height_pixels, self.output_width_pixels
                )
                if xr_index == 0:  # Only for the first one
                    _extra_time_dim = (
                        "target_time_utc" if "target_time_utc" in xr_data.dims else "time_utc"
                    )
                    # Add in time features for each timestep
                    time_image = _create_time_image(
                        xr_center,
                        time_dim=_extra_time_dim,
                        output_height_pixels=self.output_height_pixels,
                        output_width_pixels=self.output_width_pixels,
                    )
                    contexts.append(time_image)
                    # Need to add sun features
                    if self.add_sun_features:
                        sun_image = _create_sun_image(
                            image_xr=xr_center,
                            x_dim="x_osgb" if "x_osgb" in xr_data.dims else "x_geostationary",
                            y_dim="y_osgb" if "y_osgb" in xr_data.dims else "y_geostationary",
                            time_dim=_extra_time_dim,
                            normalize=True,
                        )
                        contexts.append(sun_image)
                xr_context = _resample_to_pixel_size(
                    xr_context, self.output_height_pixels, self.output_width_pixels
                )
                xr_center = xr_center.to_numpy()
                xr_context = xr_context.to_numpy()
                if len(xr_center.shape) == 3:  # Need to add channel dimension
                    xr_center = np.expand_dims(xr_center, axis=1)
                    xr_context = np.expand_dims(xr_context, axis=1)
                centers.append(xr_center)
                contexts.append(xr_context)
            # Pad out time dimension to be the same, using the largest one
            # All should have 4 dimensions at this point
            max_time_len = max(
                np.max([c.shape[0] for c in centers]), np.max([c.shape[0] for c in contexts])
            )
            for i in range(len(centers)):
                centers[i] = np.pad(
                    centers[i],
                    pad_width=((0, max_time_len - centers[i].shape[0]), (0, 0), (0, 0), (0, 0),),
                    mode="constant",
                    constant_values=0.0,
                )
            for i in range(len(contexts)):
                contexts[i] = np.pad(
                    contexts[i],
                    pad_width=((0, max_time_len - contexts[i].shape[0]), (0, 0), (0, 0), (0, 0),),
                    mode="constant",
                    constant_values=0.0,
                )
            stacked_data = np.concatenate([*centers, *contexts], axis=1)
            yield stacked_data


def _get_spatial_crop(xr_data, location, roi_height_meters: int, roi_width_meters: int):
    # Compute the index for left and right:
    half_height = roi_height_meters // 2
    half_width = roi_width_meters // 2

    left = location.x - half_width
    right = location.x + half_width
    bottom = location.y - half_height
    top = location.y + half_height
    if "x_geostationary" in xr_data.coords:
        # Convert to geostationary edges
        _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
        left, bottom = _osgb_to_geostationary(xx=left, yy=bottom)
        right, top = _osgb_to_geostationary(xx=right, yy=top)
        x_mask = (left <= xr_data.x_geostationary) & (xr_data.x_geostationary <= right)
        y_mask = (xr_data.y_geostationary <= top) & (  # Y is flipped
            bottom <= xr_data.y_geostationary
        )
        selected = xr_data.isel(x_geostationary=x_mask, y_geostationary=y_mask)
    elif "x" in xr_data.coords:
        _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
        left, bottom = _osgb_to_geostationary(xx=left, yy=bottom)
        right, top = _osgb_to_geostationary(xx=right, yy=top)
        x_mask = (left <= xr_data.x) & (xr_data.x <= right)
        y_mask = (xr_data.y <= top) & (bottom <= xr_data.y)  # Y is flipped
        selected = xr_data.isel(x=x_mask, y=y_mask)
    else:
        # Select data in the region of interest:
        x_mask = (left <= xr_data.x_osgb) & (xr_data.x_osgb <= right)
        y_mask = (xr_data.y_osgb <= top) & (bottom <= xr_data.y_osgb)
        selected = xr_data.isel(x_osgb=x_mask, y_osgb=y_mask)

    return selected


def _resample_to_pixel_size(xr_data, height_pixels, width_pixels) -> np.ndarray:
    if "x_geostationary" in xr_data.dims:
        x_coords = xr_data["x_geostationary"].values
        y_coords = xr_data["y_geostationary"].values
    elif "x" in xr_data.dims:
        x_coords = xr_data["x"].values
        y_coords = xr_data["y"].values
    else:
        x_coords = xr_data["x_osgb"].values
        y_coords = xr_data["y_osgb"].values
    # Resample down to the number of pixels wanted
    x_coords = np.linspace(x_coords[0], x_coords[-1], num=width_pixels)
    y_coords = np.linspace(y_coords[0], y_coords[-1], num=height_pixels)
    if "x_geostationary" in xr_data.dims:
        xr_data = xr_data.interp(
            x_geostationary=x_coords, y_geostationary=y_coords, method="linear"
        )
    elif "x" in xr_data.dims:
        xr_data = xr_data.interp(x=x_coords, y=y_coords, method="linear")
    else:
        xr_data = xr_data.interp(x_osgb=x_coords, y_osgb=y_coords, method="linear")
    # Extract just the data now
    return xr_data


def _create_time_image(xr_data, time_dim: str, output_height_pixels: int, output_width_pixels: int):
    times = pd.DatetimeIndex(xr_data[time_dim].values)
    # Sin+Cos of month, day, hour, after dividing by 12, 366, and 24 respecitvely
    months = times.month
    days = times.day
    hours = times.hour
    months /= 12
    days /= 366
    hours /= 24

    months = np.expand_dims(months, axis=[1, 2, 3])
    days = np.expand_dims(days, axis=[1, 2, 3])
    hours = np.expand_dims(hours, axis=[1, 2, 3])

    tile_reps = (
        1,
        output_height_pixels,
        output_width_pixels,
    )  # Has time dimension, so tile other ones by
    sin_months = np.tile(np.sin(months), reps=tile_reps)
    cos_months = np.tile(np.cos(months), reps=tile_reps)
    sin_hours = np.tile(np.sin(hours), reps=tile_reps)
    cos_hours = np.tile(np.cos(hours), reps=tile_reps)
    sin_days = np.tile(np.sin(days), reps=tile_reps)
    cos_days = np.tile(np.cos(days), reps=tile_reps)

    return np.concatenate(
        [sin_months, cos_months, sin_days, cos_days, sin_hours, cos_hours], axis=1
    )


def _create_sun_image(image_xr, x_dim, y_dim, time_dim, normalize):
    # Create empty image to use for the PV Systems, assumes image has x and y coordinates
    sun_image = np.zeros(
        (
            2,  # Azimuth and elevation
            len(image_xr[y_dim]),
            len(image_xr[x_dim]),
            len(image_xr[time_dim]),
        ),
        dtype=np.float32,
    )
    if "geostationary" in x_dim:
        transform_to_latlon = load_geostationary_area_definition_and_transform_to_latlon(image_xr)
        lats, lons = transform_to_latlon(xx=image_xr[x_dim].values, yy=image_xr[y_dim].values)
    else:
        transform_to_latlon = osgb_to_lat_lon
        lats, lons = transform_to_latlon(x=image_xr.x_osgb.values, y=image_xr.y_osgb.values)
    time_utc = image_xr[time_dim].values

    # Loop round each example to get the Sun's elevation and azimuth:
    # Go through each time on its own, lat lons still in order of image
    # TODO Make this faster
    # dt = pd.DatetimeIndex(dt)  # pvlib expects a `pd.DatetimeIndex`.
    for example_idx, (lat, lon) in enumerate(zip(lats, lons)):
        solpos = pvlib.solarposition.get_solarposition(
            time=time_utc,
            latitude=lat,
            longitude=lon,
            # Which `method` to use?
            # pyephem seemed to be a good mix between speed and ease but causes segfaults!
            # nrel_numba doesn't work when using multiple worker processes.
            # nrel_c is probably fastest but requires C code to be manually compiled:
            # https://midcdmz.nrel.gov/spa/
        )
        sun_image[0][:][example_idx] = solpos["azimuth"]
        sun_image[1][example_idx][:] = solpos["elevation"]

    # Flip back to normal ordering
    sun_image = np.transpose(sun_image, [3, 0, 1, 2])

    # Normalize.
    if normalize:
        sun_image[:, 0] = (sun_image[:, 0] - AZIMUTH_MEAN) / AZIMUTH_STD
        sun_image[:, 1] = (sun_image[:, 1] - ELEVATION_MEAN) / ELEVATION_STD
    return sun_image



"""Datapipes to check equalities"""
import logging
from typing import Optional, Union

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("check_greater_than_or_equal_to")
class CheckGreaterThanOrEqualToIterDataPipe(IterDataPipe):
    """Check greater than or equal to"""

    def __init__(
        self, source_datapipe: IterDataPipe, min_value: int, dataset_name: Optional[str] = None
    ):
        """
        Check greater than or equal to check

        Args:
            source_datapipe: Datapipe emitting Xarray object
            min_value: Minimum value
            dataset_name: Optional dataset name if checking a subset
        """
        self.source_datapipe = source_datapipe
        self.min_value = min_value
        self.dataset_name = dataset_name

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Check equality"""
        for xr_data in self.source_datapipe:
            if self.dataset_name is not None:
                check_dataset_greater_than_or_equal_to(
                    xr_data[self.dataset_name], min_value=self.min_value
                )
            else:
                check_dataset_greater_than_or_equal_to(xr_data, min_value=self.min_value)
            yield xr_data


@functional_datapipe("check_less_than_or_equal_to")
class CheckLessThanOrEqualToIterDataPipe(IterDataPipe):
    """Check less than or equal to equality"""

    def __init__(
        self, source_datapipe: IterDataPipe, max_value: int, dataset_name: Optional[str] = None
    ):
        """
        Check less than or equal to equality

        Args:
            source_datapipe: Datapipe emitting Xarray object
            max_value: Max value
            dataset_name: Optioanl dataset name if checking a subset
        """
        self.source_datapipe = source_datapipe
        self.max_value = max_value
        self.dataset_name = dataset_name

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Check equality"""
        for xr_data in self.source_datapipe:
            if self.dataset_name is not None:
                check_dataset_less_than_or_equal_to(
                    xr_data[self.dataset_name], max_value=self.max_value
                )
            else:
                check_dataset_less_than_or_equal_to(xr_data, max_value=self.max_value)
            yield xr_data


@functional_datapipe("check_not_equal_to")
class CheckNotEqualToIterDataPipe(IterDataPipe):
    """Check not equal to equality"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        value: int,
        dataset_name: Optional[str] = None,
        raise_error: bool = True,
    ):
        """
        Checks not equal to equality on the data

        Args:
            source_datapipe: Datapipe emitting Xarray object
            value: Value to check
            dataset_name: Optional dataset name if checkinga subset
            raise_error: Whether to raise an error or not
        """
        self.source_datapipe = source_datapipe
        self.value = value
        self.dataset_name = dataset_name
        self.raise_error = raise_error

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Check not equal equality"""
        for xr_data in self.source_datapipe:
            if self.dataset_name is not None:
                check_dataset_not_equal(
                    xr_data[self.dataset_name], value=self.value, raise_error=self.raise_error
                )
            else:
                check_dataset_not_equal(xr_data, value=self.value, raise_error=self.raise_error)
            yield xr_data


def check_dataset_greater_than_or_equal_to(data: xr.Dataset, min_value: int):
    """Check data is greater than a certain value"""
    if (data < min_value).any():
        message = f"Some data values are less than {min_value}. "
        message += f"The minimum value is {data.min()}. "
        raise Exception(message)


def check_dataset_less_than_or_equal_to(data: xr.Dataset, max_value: int):
    """Check data is less than a certain value"""
    if (data > max_value).any():
        message = f"Some data values are more than {max_value}"
        message += f"The maximum value  is {data.max()}. "
        raise Exception(message)


def check_dataset_not_equal(data: xr.Dataset, value: int, raise_error: bool = True):
    """Check data is not equal than a certain value"""
    if np.isclose(data, value).any():
        message = f"Some data values are equal to {value}"
        if raise_error:
            logger.error(message)
            raise Exception(message)
        else:
            logger.warning(message)
